/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <cstdlib>
#include <algorithm>
#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include <utility>
#include <vector>
//#include <futility>

#include "SDL.h"

#include "draw.hh"
#include "error.hh"
#include "point.hh"

using std::cerr;
using std::endl;
using std::max;
using std::min;
using std::map;
using std::pair;
using std::vector;

int global_zoom= 1;
Uint32 background_color= 0;
point screen_offset;
SDL_Surface* real_screen;
int width= DEFAULT_WIDTH, height= DEFAULT_HEIGHT;
map<point, point> update_rect_list;
unsigned int current_update_area= 0;
map<point, Uint32> drawn_pixels;

SDL_Surface* screen;
bool procrastinate_drawing= false;
unsigned int critical_update_area_mass= 97515;
/* Magic number, bonus points to whoever figures out what picture I got it from. */
int update_frequency= 0;

/* Prototypes for internal functions, functions should be in the same order in the
   file as in this list. */
void actually_draw();
void add_draw_point(point pos, Uint32 color);
void add_update_rect(point upper_left, point width_and_height);
void checked_draw_pixel(point pos, Uint32 color);
void checked_update_rect(point upper_left, point width_and_height);
Uint32 hilight_color(Uint32 color, unsigned int step_per_channel);
bool pos_off_screen(point pos);
bool rect_off_screen(point upper_left, point lower_right);
void unchecked_draw_pixel(point pos, Uint32 color);
void draw_hollow_square(point ul, point lr, Uint32 color);

void actually_draw()
  {
    for(map<point, Uint32>::iterator pix= drawn_pixels.begin(); pix != drawn_pixels.end(); ++pix)
    {
      unchecked_draw_pixel(pix->first, pix->second);
    }
  }

void add_draw_point(point pos, Uint32 color)
  {
    drawn_pixels[pos]= color;
  }

void add_update_rect(point upper_left, point width_and_height)
  {
//     map<point, point>::iterator rect= update_rect_list.find(upper_left);
//     if(rect == update_rect_list.end())
//     {
      update_rect_list[upper_left]= width_and_height;
//       return;
//     }
//     rect->second= max(width_and_height, rect->second);
      current_update_area+= width_and_height.x * width_and_height.y;
  }

void checked_draw_pixel(point pos, Uint32 color)
  {
    if(!pos_off_screen(pos))
    {
//       cerr << '.';
      if(procrastinate_drawing)
      {
        add_draw_point(pos, color);
      }
      else
      {
        unchecked_draw_pixel(pos, color);
      }
    }
  }

void checked_update_rect(point upper_left, point width_and_height)
  {
    if(!rect_off_screen(upper_left, upper_left + width_and_height))
    {
      point real_ul(max(upper_left.x, 0), max(upper_left.y, 0));
      point real_wah(min(width_and_height.x, screen->w - upper_left.x)
        , min(width_and_height.y, screen->h - upper_left.y));
      add_update_rect(real_ul, real_wah);
    }
  }

Uint32 hilight_color(Uint32 color, unsigned int step_per_channel)
  {
    if(!color) return color; /* Refuse to hilight black. */
    if(color == 0xffffff) return color; /* Or white. */
    step_per_channel&= 0xff;
    return (color | (step_per_channel) | ((step_per_channel) << 8)
      | ((step_per_channel) << 16)) & 0xffffff;
  }

bool pos_off_screen(point pos)
  {
    return pos.x < 0 || pos.x > screen->w - 1 || pos.y < 0  || pos.y > screen->h - 1;
  }

bool rect_off_screen(point upper_left, point lower_right)
  {
    return (lower_right.x < 0) || (upper_left.x >= screen->w)
      || (lower_right.y < 0) || (upper_left.y >= screen->h);
  }

void unchecked_draw_pixel(point pos, Uint32 color)
  {
    (*((Uint32*)((Uint8*)(screen->pixels) + ((pos.y)* screen->pitch)) + (pos.x)))= color;
  }

/* End of internal functions. */

void set_screen_geometry(int new_width, int new_height)
  {
    width= new_width;
    height= new_height;
  }

void init_drawing() throw(error)
  {
    std::string foo; /* I don't know why it's necessary, but it is.  Probably
            specifies special attributes of the string the ostringstream contains. */
    if(SDL_Init(SDL_INIT_VIDEO) == -1)
    {
      std::ostringstream err_out(foo);
      err_out << "Could not initialize SDL:  " << SDL_GetError();
      throw error(err_out.str());
    }
    real_screen= SDL_SetVideoMode(width, height, 32, SDL_HWSURFACE | SDL_NOFRAME);
    SDL_SetClipRect(real_screen, NULL);
    if(real_screen == NULL)
    {
      std::ostringstream err_out(foo);
      err_out << "Couldn't set " << width << 'x' << height << " 32 bpp video mode:  " << SDL_GetError();
      throw error(err_out.str());
    }
    SDL_WM_SetCaption("Kyzentun's Drawing Tool", "KDT");
//     screen= SDL_CreateRGBSurface(SDL_SWSURFACE, real_screen->w, real_screen->h
//       , 32, real_screen->format->Rmask, real_screen->format->Gmask
//       , real_screen->format->Bmask, real_screen->format->Amask);
//     SDL_SetClipRect(screen, NULL);
//     if(screen == NULL)
//     {
//       std::ostringstream err_out(foo);
//       err_out << "Couldn't set up shadow screen." << endl;
//       throw error(err_out.str());
//     }
    screen= real_screen;
    screen_offset.x= screen->w / 2;
    screen_offset.y= screen->h / 2;
    black_screen();
  }

void reset_drawing()
  {
    SDL_FillRect(screen, NULL, 0);
    SDL_Flip(screen);
    screen_offset.x= screen->w / 2;
    screen_offset.y= screen->h / 2;
  }

void frequency_update_drawn_rects(int fi)
  {
    if(update_frequency > fi)
    {
      update_drawn_rects();
    }
  }

void update_drawn_rects()
  {
    if(procrastinate_drawing)
    {
      actually_draw();
      drawn_pixels.clear();
    }
    if(current_update_area < critical_update_area_mass)
    {
      for(map<point, point>::iterator rect= update_rect_list.begin(); rect != update_rect_list.end(); ++rect)
      {
//        SDL_Rect src_rect= {rect->first.x, rect->first.y, rect->second.x, rect->second.y};
//        SDL_BlitSurface(screen, &src_rect, real_screen, &src_rect);
        SDL_UpdateRect(real_screen, rect->first.x, rect->first.y, rect->second.x, rect->second.y);
      }
    }
    else
    {
//       cerr << "Critical update mass surpassed, flipping screen." << endl;
      SDL_Flip(screen);
    }
    current_update_area= 0;
    update_rect_list.clear();
  }

void draw_hollow_square(point const ul, point const lr, Uint32 color)
  {
    point curr_pix= ul;
    checked_draw_pixel(curr_pix, color);
    for(; curr_pix.x < lr.x; ++curr_pix.x)
    {
      checked_draw_pixel(curr_pix, color);
    }
    --curr_pix.x;
    for(; curr_pix.y < lr.y; ++curr_pix.y)
    {
      checked_draw_pixel(curr_pix, color);
    }
    --curr_pix.y;
    for(; curr_pix.x >= ul.x; --curr_pix.x)
    {
      checked_draw_pixel(curr_pix, color);
    }
    ++curr_pix.x;
    for(; curr_pix.y >= ul.y; --curr_pix.y)
    {
      checked_draw_pixel(curr_pix, color);
    }
  }

void draw_pixel(point pos, Uint32 color, bool hilight)
  {
//     cerr << '(' << pos.x << ", " << pos.y << ')';
    pos+= screen_offset / global_zoom;
    pos*= global_zoom;
    if(rect_off_screen(pos, pos + global_zoom))
    {
      return;
    }
//     cerr << '.';
    if(global_zoom == 1)
    {
      if(hilight)
      {
        color= hilight_color(color, 64);
      }
      checked_draw_pixel(pos, color);
      checked_update_rect(pos, point(global_zoom, global_zoom));
      return;
    }
    if(global_zoom == 2)
    {
      point d_pos= pos;
      checked_draw_pixel(d_pos, color);
      ++d_pos.x; ++d_pos.y;
      checked_draw_pixel(d_pos, color);
      if(hilight)
      {
        color= hilight_color(color, 64);
      }
      --d_pos.x;
      checked_draw_pixel(d_pos, color);
      ++d_pos.x; --d_pos.y;
      checked_draw_pixel(d_pos, color);
      checked_update_rect(pos, point(2, 2));
      return;
    }
    point p_min= pos;
    point p_max= p_min + (global_zoom);
    if(!hilight)
    {
      point d_pos= pos;
      for(d_pos.y= p_min.y; d_pos.y < p_max.y; ++d_pos.y)
      {
        for(d_pos.x= p_min.x; d_pos.x < p_max.x; ++d_pos.x)
        {
          checked_draw_pixel(d_pos, color);
        }
      }
      checked_update_rect(p_min, p_max - p_min);
      return;
    }
    int num_tiers= global_zoom / 2 + (global_zoom % 2);
    for(int t= 0; t < num_tiers; ++t)
    {
      color= hilight_color(color, (64 / num_tiers) * t);
      draw_hollow_square(p_min + t, p_max - t, color);
    }
    checked_update_rect(p_min, p_max - p_min);
  }

void draw_unzoomed_rect(point upper_left, point width_and_height, Uint32 color)
  {
    upper_left*= global_zoom;
    upper_left+= screen_offset;
    if(rect_off_screen(upper_left, upper_left + width_and_height))
    {
      return;
    }
    if(width_and_height.x < 1 || width_and_height.y < 1)
    { /* I refuse to draw a rect of zero or negative width or height. */
      return;
    }
    point current_point= upper_left;
    for(current_point.y= upper_left.y; current_point.y < upper_left.y + width_and_height.y; ++current_point.y)
    {
      for(current_point.x= upper_left.x; current_point.x < upper_left.x + width_and_height.x; ++current_point.x)
      {
        checked_draw_pixel(current_point, color);
      }
    }
    checked_update_rect(upper_left, width_and_height);
  }

void black_screen()
  {
    SDL_FillRect(screen, NULL, background_color);
    SDL_Flip(screen);
  }

Uint32 color_at_point(point p)
  {
    return (*((Uint32*)((Uint8*)(screen->pixels) + ((p.y)* screen->pitch)) + (p.x)));
  }

Uint32 interpolate(Uint32 s_color, Uint32 c_color, Uint32 e_color, int line_length)
  {
    /* Uint32 diff1= ((s_color >> 16) & 255) - ((e_color >> 16) & 255);
    Uint32 diff2= ((s_color >> 8) & 255) - ((e_color >> 8) & 255);
    Uint32 diff3= ((s_color) & 255) - ((e_color) & 255);
    Uint32 ret= s_color + (((diff1 / line_length) << 16)
      | ((diff2 / line_length) << 8) | (diff3 / line_length));
      return ret; */
    return c_color +
      ((((((s_color >> 16) & 255) - ((e_color >> 16) & 255)) / line_length) << 16)
      | (((((s_color >> 8) & 255) - ((e_color >> 8) & 255)) / line_length) << 8)
        | (((s_color & 255) - (e_color & 255)) / line_length));
  }

unsigned int increase_zoom()
  {
    ++global_zoom;
    global_zoom&= 31;
    if(!global_zoom)
    {
      global_zoom= 31;
    }
    return global_zoom;
  }

unsigned int decrease_zoom()
  {
    --global_zoom;
    if(!global_zoom)
    {
      global_zoom= 1;
    }
    return global_zoom;
  }

void set_zoom(unsigned int new_zoom)
  {
    global_zoom= new_zoom & 31;
    if(global_zoom == 0)
    {
      global_zoom= 1;
    }
  }

point get_screen_offset()
  {
    return screen_offset;
  }

int move_horizontally(int amount)
  {
    screen_offset.x+= amount;
    SDL_FillRect(screen, NULL, 0);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    return screen_offset.x;
  }

int move_vertically(int amount)
  {
    screen_offset.y+= amount;
    SDL_FillRect(screen, NULL, 0);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    return screen_offset.y;
  }

void set_background_color(Uint32 new_color)
  {
    background_color= new_color;
  }

Uint32 get_background_color()
  {
    return background_color;
  }
