/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <fstream>
#include <map>

#include "draw.hh"
#include "environment.hh"
#include "png_save.hh"
#include "sdl_command_line.hh"
#include "str_tools.hh"

using std::list;
using std::string;
using std::ifstream;
using std::ofstream;
using std::map;

#include <iostream>
using std::cerr;
using std::endl;

environment::environment()
  :next_op_multiplier(1, 1)
   , swap_r_l(false)
   , is_sub_env(false)
   , quit(false)
  {}

void environment::process_sdlk(SDLKey command)
  {
    lossage.push_back(command);
    if(lossage.size() > 100)
    {
      lossage.pop_front();
    }
    int temp= convert_sdlk_to_int(command);
    if(temp != -1)
    {
      happy_op_env.add_op_to_current_list(temp);
    }
    if(!multiplier_op(convert_sdlk_to_int(command)))
    {
      while(next_op_multiplier.back() > 1)
      {
        try
        {
          interpret_sdlk(command);
        }
        catch(error e)
        {
          display_error(e.cause);
        }
        --next_op_multiplier.back();
        frequency_update_drawn_rects(2);
      }
      try
      {
        interpret_sdlk(command);
      }
      catch(error e)
      {
        display_error(e.cause);
      }
    }
    else
    {
      try
      {
        interpret_sdlk(command);
      }
      catch(error e)
      {
        display_error(e.cause);
      }
    }
    frequency_update_drawn_rects(-2);
  }

void environment::process_int(int command)
  {
    happy_op_env.add_op_to_current_list(command);
    if(!multiplier_op(command))
    {
      while(next_op_multiplier.back() > 1)
      {
        try
        {
          interpret_int(command);
        }
        catch(error e)
        {
          display_error(e.cause);
        }
        --next_op_multiplier.back();
        frequency_update_drawn_rects(2);
      }
      try
      {
        interpret_int(command);
      }
      catch(error e)
      {
        display_error(e.cause);
      }
    }
    else
    {
      try
      {
        interpret_int(command);
      }
      catch(error e)
      {
        display_error(e.cause);
      }
    }
    frequency_update_drawn_rects(1);
  }

void environment::process_command_line(std::list<std::string> command)
  {
    /* CHANGEME:  Add up a command line interface for environment.
       Only pass the command to happy_line_env if that's where it's supposed to go. */
    try
    {
      if(command.front() == "wtf?")
      {
        view_lossage(*this);
      }
      happy_line_env.parse_top_level_command(command);
    }
    catch(error e)
    {
      display_error(e.cause);
    }
  }

void environment::process_int_script(std::string file_name) throw(error)
  {
    bool prev_sub_env= is_sub_env;
    is_sub_env= true;
    next_op_multiplier.push_back(1);
    ifstream in_file(file_name.c_str());
    if(!in_file)
    {
      throw error("Invalid file:  " + file_name);
    }
    string s;
    while(getline(in_file, s))
    {
      if(s == "[ops]")
      {
        break;
      }
    }
    script= &in_file;
    int command;
    while(in_file >> command)
    {
      process_int(command);
    }
    in_file.close();
    next_op_multiplier.pop_back();
    is_sub_env= prev_sub_env;
  }

void environment::process_int_rlswap_script(std::string file_name) throw(error)
  {
    bool prev_sub_env= is_sub_env;
    is_sub_env= true;
    next_op_multiplier.push_back(1);
    ifstream in_file(file_name.c_str());
    if(!in_file)
    {
      throw error("Invalid file:  " + file_name);
    }
    string s;
    while(getline(in_file, s))
    {
      if(s == "[ops]")
      {
        break;
      }
    }
    script= &in_file;
    int command;
    while(in_file >> command)
    {
      if(command == turn_left_int)
      {
        command= turn_right_int;
      }
      else if(command == turn_right_int)
      {
        command= turn_left_int;
      }
      process_int(command);
    }
    in_file.close();
    next_op_multiplier.pop_back();
    is_sub_env= prev_sub_env;
  }

void environment::process_command_script(std::string file_name)
  {
    bool prev_sub_env= is_sub_env;
    is_sub_env= true;
    next_op_multiplier.push_back(1);
    ifstream in_file(file_name.c_str());
    if(!in_file)
    {
      throw error("Invalid file:  " + file_name);
    }
    string s;
    while(getline(in_file, s))
    {
      if(s == "[commands]")
      {
        break;
      }
    }
    script= &in_file;
    while(getline(in_file, s))
    {
      list<string> command;
      cook_command_line(s, command);
      process_command_line(command);
    }
    in_file.close();
    next_op_multiplier.pop_back();
    is_sub_env= prev_sub_env;
  }

void environment::interpret_int(int command)
  {
    static map<int, void(*)(environment&)> int_to_op_func_ptr;
    static bool map_inited= false;
    if(!map_inited)
    {
      int_to_op_func_ptr[set_and_move_int]= &set_and_move;
      int_to_op_func_ptr[set_without_moving_int]= &set_without_moving;
      int_to_op_func_ptr[move_without_setting_int]= &move_without_setting;
      int_to_op_func_ptr[turn_left_int]= &turn_left;
      int_to_op_func_ptr[turn_right_int]= &turn_right;
      int_to_op_func_ptr[trip_int]= &trip_next;
      int_to_op_func_ptr[quad_int]= &quad_next;
      int_to_op_func_ptr[quin_int]= &quin_next;
      int_to_op_func_ptr[turn_to_first_int]= &turn_to_first;
      int_to_op_func_ptr[turn_to_last_int]= &turn_to_last;
      int_to_op_func_ptr[turn_to_indexed_int]= &turn_to_indexed;
      int_to_op_func_ptr[meta_turn_left_int]= &meta_turn_left;
      int_to_op_func_ptr[meta_turn_right_int]= &meta_turn_right;
      int_to_op_func_ptr[meta_turn_to_first_int]= &meta_turn_to_first;
      int_to_op_func_ptr[meta_turn_to_last_int]= &meta_turn_to_last;
      int_to_op_func_ptr[meta_turn_to_indexed_int]= &meta_turn_to_indexed;
      int_to_op_func_ptr[inc_curr_line_red_1_int]= &inc_curr_line_red_1;
      int_to_op_func_ptr[inc_curr_line_red_4_int]= &inc_curr_line_red_4;
      int_to_op_func_ptr[inc_curr_line_red_16_int]= &inc_curr_line_red_16;
      int_to_op_func_ptr[inc_curr_line_red_64_int]= &inc_curr_line_red_64;
      int_to_op_func_ptr[dec_curr_line_red_1_int]= &dec_curr_line_red_1;
      int_to_op_func_ptr[dec_curr_line_red_4_int]= &dec_curr_line_red_4;
      int_to_op_func_ptr[dec_curr_line_red_16_int]= &dec_curr_line_red_16;
      int_to_op_func_ptr[dec_curr_line_red_64_int]= &dec_curr_line_red_64;
      int_to_op_func_ptr[zero_curr_line_red_int]= &zero_curr_line_red;
      int_to_op_func_ptr[max_curr_line_red_int]= &max_curr_line_red;
      int_to_op_func_ptr[inc_curr_line_green_1_int]= &inc_curr_line_green_1;
      int_to_op_func_ptr[inc_curr_line_green_4_int]= &inc_curr_line_green_4;
      int_to_op_func_ptr[inc_curr_line_green_16_int]= &inc_curr_line_green_16;
      int_to_op_func_ptr[inc_curr_line_green_64_int]= &inc_curr_line_green_64;
      int_to_op_func_ptr[dec_curr_line_green_1_int]= &dec_curr_line_green_1;
      int_to_op_func_ptr[dec_curr_line_green_4_int]= &dec_curr_line_green_4;
      int_to_op_func_ptr[dec_curr_line_green_16_int]= &dec_curr_line_green_16;
      int_to_op_func_ptr[dec_curr_line_green_64_int]= &dec_curr_line_green_64;
      int_to_op_func_ptr[zero_curr_line_green_int]= &zero_curr_line_green;
      int_to_op_func_ptr[max_curr_line_green_int]= &max_curr_line_green;
      int_to_op_func_ptr[inc_curr_line_blue_1_int]= &inc_curr_line_blue_1;
      int_to_op_func_ptr[inc_curr_line_blue_4_int]= &inc_curr_line_blue_4;
      int_to_op_func_ptr[inc_curr_line_blue_16_int]= &inc_curr_line_blue_16;
      int_to_op_func_ptr[inc_curr_line_blue_64_int]= &inc_curr_line_blue_64;
      int_to_op_func_ptr[dec_curr_line_blue_1_int]= &dec_curr_line_blue_1;
      int_to_op_func_ptr[dec_curr_line_blue_4_int]= &dec_curr_line_blue_4;
      int_to_op_func_ptr[dec_curr_line_blue_16_int]= &dec_curr_line_blue_16;
      int_to_op_func_ptr[dec_curr_line_blue_64_int]= &dec_curr_line_blue_64;
      int_to_op_func_ptr[zero_curr_line_blue_int]= &zero_curr_line_blue;
      int_to_op_func_ptr[max_curr_line_blue_int]= &max_curr_line_blue;
//       int_to_op_func_ptr[inc_zoom_int]= &inc_zoom;
//       int_to_op_func_ptr[dec_zoom_int]= &dec_zoom;
//       int_to_op_func_ptr[move_center_up_int]= &move_center_up;
//       int_to_op_func_ptr[move_center_down_int]= &move_center_down;
//       int_to_op_func_ptr[move_center_left_int]= &move_center_left;
//       int_to_op_func_ptr[move_center_right_int]= &move_center_right;
//       int_to_op_func_ptr[save_int_script_int]= &save_int_script;
//       int_to_op_func_ptr[use_command_line_int]= &use_command_line;
//       int_to_op_func_ptr[load_int_script_int]= &load_int_script;
//       int_to_op_func_ptr[load_command_script_int]= &load_command_script;
//       int_to_op_func_ptr[load_new_template_set_int]= &load_new_template_set;
//       int_to_op_func_ptr[drop_curr_template_set_int]= &drop_curr_template_set;
   }
    map<int, void(*)(environment&)>::iterator fp= int_to_op_func_ptr.find(command);
    if(fp != int_to_op_func_ptr.end())
    {
      fp->second(*this);
    }
  }

void environment::interpret_sdlk(SDLKey command)
  {
    static map<SDLKey, void(*)(environment&)> sdlk_to_op_func_ptr;
    static bool map_inited= false;
    if(!map_inited)
    {
      sdlk_to_op_func_ptr[SDLK_KP1]= &set_and_move;
      sdlk_to_op_func_ptr[SDLK_KP2]= &set_without_moving;
      sdlk_to_op_func_ptr[SDLK_KP3]= &move_without_setting;
      if(!swap_r_l)
      {
        sdlk_to_op_func_ptr[SDLK_KP4]= &turn_left;
        sdlk_to_op_func_ptr[SDLK_KP6]= &turn_right;
      }
      else
      {
        sdlk_to_op_func_ptr[SDLK_KP6]= &turn_left;
        sdlk_to_op_func_ptr[SDLK_KP4]= &turn_right;
      }
      sdlk_to_op_func_ptr[SDLK_KP7]= &trip_next;
      sdlk_to_op_func_ptr[SDLK_KP8]= &quad_next;
      sdlk_to_op_func_ptr[SDLK_KP9]= &quin_next;
      sdlk_to_op_func_ptr[SDLK_KP_DIVIDE]= &turn_to_first;
      sdlk_to_op_func_ptr[SDLK_KP_MINUS]= &turn_to_last;
      sdlk_to_op_func_ptr[SDLK_KP_MULTIPLY]= &turn_to_indexed;
      sdlk_to_op_func_ptr[SDLK_x]= &meta_turn_left;
      sdlk_to_op_func_ptr[SDLK_v]= &meta_turn_right;
      sdlk_to_op_func_ptr[SDLK_z]= &meta_turn_to_first;
      sdlk_to_op_func_ptr[SDLK_b]= &meta_turn_to_last;
      sdlk_to_op_func_ptr[SDLK_c]= &meta_turn_to_indexed;
      sdlk_to_op_func_ptr[SDLK_6]= &inc_curr_line_red_1;
      sdlk_to_op_func_ptr[SDLK_7]= &inc_curr_line_red_4;
      sdlk_to_op_func_ptr[SDLK_8]= &inc_curr_line_red_16;
      sdlk_to_op_func_ptr[SDLK_9]= &inc_curr_line_red_64;
      sdlk_to_op_func_ptr[SDLK_5]= &dec_curr_line_red_1;
      sdlk_to_op_func_ptr[SDLK_4]= &dec_curr_line_red_4;
      sdlk_to_op_func_ptr[SDLK_3]= &dec_curr_line_red_16;
      sdlk_to_op_func_ptr[SDLK_2]= &dec_curr_line_red_64;
      sdlk_to_op_func_ptr[SDLK_1]= &zero_curr_line_red;
      sdlk_to_op_func_ptr[SDLK_0]= &max_curr_line_red;
      sdlk_to_op_func_ptr[SDLK_y]= &inc_curr_line_green_1;
      sdlk_to_op_func_ptr[SDLK_u]= &inc_curr_line_green_4;
      sdlk_to_op_func_ptr[SDLK_i]= &inc_curr_line_green_16;
      sdlk_to_op_func_ptr[SDLK_o]= &inc_curr_line_green_64;
      sdlk_to_op_func_ptr[SDLK_t]= &dec_curr_line_green_1;
      sdlk_to_op_func_ptr[SDLK_r]= &dec_curr_line_green_4;
      sdlk_to_op_func_ptr[SDLK_e]= &dec_curr_line_green_16;
      sdlk_to_op_func_ptr[SDLK_w]= &dec_curr_line_green_64;
      sdlk_to_op_func_ptr[SDLK_q]= &zero_curr_line_green;
      sdlk_to_op_func_ptr[SDLK_p]= &max_curr_line_green;
      sdlk_to_op_func_ptr[SDLK_h]= &inc_curr_line_blue_1;
      sdlk_to_op_func_ptr[SDLK_j]= &inc_curr_line_blue_4;
      sdlk_to_op_func_ptr[SDLK_k]= &inc_curr_line_blue_16;
      sdlk_to_op_func_ptr[SDLK_l]= &inc_curr_line_blue_64;
      sdlk_to_op_func_ptr[SDLK_g]= &dec_curr_line_blue_1;
      sdlk_to_op_func_ptr[SDLK_f]= &dec_curr_line_blue_4;
      sdlk_to_op_func_ptr[SDLK_d]= &dec_curr_line_blue_16;
      sdlk_to_op_func_ptr[SDLK_s]= &dec_curr_line_blue_64;
      sdlk_to_op_func_ptr[SDLK_a]= &zero_curr_line_blue;
      sdlk_to_op_func_ptr[SDLK_SEMICOLON]= &max_curr_line_blue;
      sdlk_to_op_func_ptr[SDLK_EQUALS]= &inc_zoom;
      sdlk_to_op_func_ptr[SDLK_MINUS]= &dec_zoom;
      sdlk_to_op_func_ptr[SDLK_UP]= &move_center_up;
      sdlk_to_op_func_ptr[SDLK_DOWN]= &move_center_down;
      sdlk_to_op_func_ptr[SDLK_LEFT]= &move_center_left;
      sdlk_to_op_func_ptr[SDLK_RIGHT]= &move_center_right;
      sdlk_to_op_func_ptr[SDLK_INSERT]= &save_int_script;
      sdlk_to_op_func_ptr[SDLK_n]= &first_op_list;
      sdlk_to_op_func_ptr[SDLK_m]= &prev_op_list;
      sdlk_to_op_func_ptr[SDLK_COMMA]= &indexed_op_list;
      sdlk_to_op_func_ptr[SDLK_PERIOD]= &next_op_list;
      sdlk_to_op_func_ptr[SDLK_SLASH]= &last_op_list;
      sdlk_to_op_func_ptr[SDLK_RIGHTBRACKET]= &pop_last_op;
      sdlk_to_op_func_ptr[SDLK_BACKSLASH]= &new_op_list;
      sdlk_to_op_func_ptr[SDLK_RETURN]= &use_command_line;
      sdlk_to_op_func_ptr[SDLK_DELETE]= &load_int_script;
      sdlk_to_op_func_ptr[SDLK_LEFTBRACKET]= &load_int_rlswap_script;
      sdlk_to_op_func_ptr[SDLK_HOME]= &load_command_script;
      sdlk_to_op_func_ptr[SDLK_PAGEUP]= &load_new_template_set;
      sdlk_to_op_func_ptr[SDLK_PAGEDOWN]= &drop_curr_template_set;
      sdlk_to_op_func_ptr[SDLK_END]= &reset;
      sdlk_to_op_func_ptr[SDLK_ESCAPE]= &leave;
      sdlk_to_op_func_ptr[SDLK_QUOTE]= &save_normal;
    }
    map<SDLKey, void(*)(environment&)>::iterator fp= sdlk_to_op_func_ptr.find(command);
    if(fp != sdlk_to_op_func_ptr.end())
    {
      fp->second(*this);
    }
  }

int environment::convert_sdlk_to_int(SDLKey command)
  {
    static map<SDLKey, int> sdlk_to_int;
    static bool map_inited= false;
    if(!map_inited)
    {
      sdlk_to_int[SDLK_KP1]= set_and_move_int;
      sdlk_to_int[SDLK_KP2]= set_without_moving_int;
      sdlk_to_int[SDLK_KP3]= move_without_setting_int;
      if(!swap_r_l)
      {
        sdlk_to_int[SDLK_KP4]= turn_left_int;
        sdlk_to_int[SDLK_KP6]= turn_right_int;
      }
      else
      {
        sdlk_to_int[SDLK_KP6]= turn_left_int;
        sdlk_to_int[SDLK_KP4]= turn_right_int;
      }
      sdlk_to_int[SDLK_KP7]= trip_int;
      sdlk_to_int[SDLK_KP8]= quad_int;
      sdlk_to_int[SDLK_KP9]= quin_int;
      sdlk_to_int[SDLK_KP_DIVIDE]= turn_to_first_int;
      sdlk_to_int[SDLK_KP_MINUS]= turn_to_last_int;
      sdlk_to_int[SDLK_KP_MULTIPLY]= turn_to_indexed_int;
      sdlk_to_int[SDLK_x]= meta_turn_left_int;
      sdlk_to_int[SDLK_v]= meta_turn_right_int;
      sdlk_to_int[SDLK_z]= meta_turn_to_first_int;
      sdlk_to_int[SDLK_b]= meta_turn_to_last_int;
      sdlk_to_int[SDLK_c]= meta_turn_to_indexed_int;
      sdlk_to_int[SDLK_6]= inc_curr_line_red_1_int;
      sdlk_to_int[SDLK_7]= inc_curr_line_red_4_int;
      sdlk_to_int[SDLK_8]= inc_curr_line_red_16_int;
      sdlk_to_int[SDLK_9]= inc_curr_line_red_64_int;
      sdlk_to_int[SDLK_5]= dec_curr_line_red_1_int;
      sdlk_to_int[SDLK_4]= dec_curr_line_red_4_int;
      sdlk_to_int[SDLK_3]= dec_curr_line_red_16_int;
      sdlk_to_int[SDLK_2]= dec_curr_line_red_64_int;
      sdlk_to_int[SDLK_1]= zero_curr_line_red_int;
      sdlk_to_int[SDLK_0]= max_curr_line_red_int;
      sdlk_to_int[SDLK_y]= inc_curr_line_green_1_int;
      sdlk_to_int[SDLK_u]= inc_curr_line_green_4_int;
      sdlk_to_int[SDLK_i]= inc_curr_line_green_16_int;
      sdlk_to_int[SDLK_o]= inc_curr_line_green_64_int;
      sdlk_to_int[SDLK_t]= dec_curr_line_green_1_int;
      sdlk_to_int[SDLK_r]= dec_curr_line_green_4_int;
      sdlk_to_int[SDLK_e]= dec_curr_line_green_16_int;
      sdlk_to_int[SDLK_w]= dec_curr_line_green_64_int;
      sdlk_to_int[SDLK_q]= zero_curr_line_green_int;
      sdlk_to_int[SDLK_p]= max_curr_line_green_int;
      sdlk_to_int[SDLK_h]= inc_curr_line_blue_1_int;
      sdlk_to_int[SDLK_j]= inc_curr_line_blue_4_int;
      sdlk_to_int[SDLK_k]= inc_curr_line_blue_16_int;
      sdlk_to_int[SDLK_l]= inc_curr_line_blue_64_int;
      sdlk_to_int[SDLK_g]= dec_curr_line_blue_1_int;
      sdlk_to_int[SDLK_f]= dec_curr_line_blue_4_int;
      sdlk_to_int[SDLK_d]= dec_curr_line_blue_16_int;
      sdlk_to_int[SDLK_s]= dec_curr_line_blue_64_int;
      sdlk_to_int[SDLK_a]= zero_curr_line_blue_int;
      sdlk_to_int[SDLK_SEMICOLON]= max_curr_line_blue_int;
//       sdlk_to_int[SDLK_EQUALS]= inc_zoom_int;
//       sdlk_to_int[SDLK_MINUS]= dec_zoom_int;
//       sdlk_to_int[SDLK_UP]= move_center_up_int;
//       sdlk_to_int[SDLK_DOWN]= move_center_down_int;
//       sdlk_to_int[SDLK_LEFT]= move_center_left_int;
//       sdlk_to_int[SDLK_RIGHT]= move_center_right_int;
//       sdlk_to_int[SDLK_INSERT]= save_int_script_int;
//       sdlk_to_int[SDLK_RETURN]= use_command_line_int;
//       sdlk_to_int[SDLK_DELETE]= load_int_script_int;
//       sdlk_to_int[SDLK_HOME]= load_command_script_int;
//       sdlk_to_int[SDLK_PAGEUP]= load_new_template_set_int;
//       sdlk_to_int[SDLK_PAGEDOWN]= drop_curr_template_set_int;
    }
    map<SDLKey, int>::iterator fint= sdlk_to_int.find(command);
    if(fint != sdlk_to_int.end())
    {
      return fint->second;
    }
    return -1;
  }

bool environment::multiplier_op(int op)
  {
    return (op == trip_int) || (op == quad_int) || (op == quin_int);
  }

void environment::mung()
  {
    happy_line_env.mung_all();
    happy_line_env.happy_template_env.mung();
    happy_op_env.mung();
    black_screen();
    happy_line_env.change_current_line(starting_point
      , happy_line_env.happy_template_env.get_current_template()
      , false, false, false, 0xffffff, 0xffffff);
    happy_line_env.draw_current_line();
  }

void environment::init(std::list<std::string>& template_file_names)
  {
    happy_op_env.init();
    happy_line_env.init(starting_point, template_file_names);
    happy_line_env.draw_current_line();
  }

environment current_env;

/* These would be members if I hadn't heard such bad things about pointers to member
   functions. */

void cond_undraw_curr(environment& happy_env)
  {
    if(!happy_env.is_sub_env)
    {
      happy_env.happy_line_env.undraw_current_line();
    }
  }

void cond_draw_curr(environment& happy_env)
  {
    if(!happy_env.is_sub_env)
    {
      happy_env.happy_line_env.draw_current_line();
    }
  }

void turn_left(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_previous_template();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void turn_right(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_next_template();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void turn_to_first(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_first_template();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void turn_to_last(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_last_template();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void turn_to_indexed(environment& happy_env)
  {
    int index;
    if(!happy_env.is_sub_env)
    {
      index= str_to_int(get_int());
    }
    else
    {
      if(!((*(happy_env.script)) >> index))
      {
        throw error("Invalid index.");
      }
    }
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_indexed_template(index);
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void meta_turn_left(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_previous_template_set();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void meta_turn_right(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_next_template_set();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void meta_turn_to_first(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_first_template_set();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void meta_turn_to_last(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_last_template_set();
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void meta_turn_to_indexed(environment& happy_env)
  {
    int index;
    if(!happy_env.is_sub_env)
    {
      index= str_to_int(get_int());
    }
    else
    {
      if(!((*(happy_env.script)) >> index))
      {
        throw error("Invalid index.");
      }
    }
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.happy_template_env.select_indexed_template_set(index);
    happy_env.happy_line_env.reset_current_line();
    cond_draw_curr(happy_env);
  }

void inc_curr_line_red_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, 1);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_red_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, 4);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_red_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, 16);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_red_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, 64);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_red_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, -1);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_red_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, -4);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_red_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, -16);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_red_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 2, -64);
    cond_draw_curr(happy_env);
  }

void zero_curr_line_red(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(0, 3);
    cond_draw_curr(happy_env);
  }

void max_curr_line_red(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(255, 3);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_green_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, 1);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_green_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, 4);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_green_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, 16);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_green_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, 64);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_green_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, -1);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_green_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, -4);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_green_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, -16);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_green_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 1, -64);
    cond_draw_curr(happy_env);
  }

void zero_curr_line_green(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(0, 4);
    cond_draw_curr(happy_env);
  }

void max_curr_line_green(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(255, 4);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_blue_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, 1);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_blue_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, 4);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_blue_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, 16);
    cond_draw_curr(happy_env);
  }

void inc_curr_line_blue_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, 64);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_blue_1(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, -1);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_blue_4(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, -4);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_blue_16(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, -16);
    cond_draw_curr(happy_env);
  }

void dec_curr_line_blue_64(environment& happy_env)
  {
    happy_env.happy_line_env.shift_current_line_color(true, 0, -64);
    cond_draw_curr(happy_env);
  }

void zero_curr_line_blue(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(0, 5);
    cond_draw_curr(happy_env);
  }

void max_curr_line_blue(environment& happy_env)
  {
    happy_env.happy_line_env.change_current_line(255, 5);
    cond_draw_curr(happy_env);
  }

void set_and_move(environment& happy_env)
  {
    happy_env.happy_line_env.add_curr_line_to_curr_set();
    happy_env.happy_line_env.change_current_line(
      happy_env.happy_line_env.current_line_end()
      , happy_env.happy_line_env.happy_template_env.get_current_template());
    happy_env.happy_line_env.draw_last_line();
    cond_draw_curr(happy_env);
  }

void set_without_moving(environment& happy_env)
  {
    happy_env.happy_line_env.add_curr_line_to_curr_set();
    happy_env.happy_line_env.draw_last_line();
    cond_draw_curr(happy_env);
  }

void move_without_setting(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    happy_env.happy_line_env.change_current_line(
      happy_env.happy_line_env.current_line_end()
      , happy_env.happy_line_env.happy_template_env.get_current_template());
    cond_draw_curr(happy_env);
  }

void trip_next(environment& happy_env)
  {
    multiply_next(happy_env, 3);
  }

void quad_next(environment& happy_env)
  {
    multiply_next(happy_env, 4);
  }

void quin_next(environment& happy_env)
  {
    multiply_next(happy_env, 5);
  }

void multiply_next(environment& happy_env, unsigned int factor)
  {
    happy_env.next_op_multiplier.back()*= factor;
  }

void inc_zoom(environment& happy_env)
  {
    black_screen();
    increase_zoom();
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void dec_zoom(environment& happy_env)
  {
    black_screen();
    decrease_zoom();
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void move_center_up(environment& happy_env)
  {
    black_screen();
    move_vertically(-32);
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void move_center_down(environment& happy_env)
  {
    black_screen();
    move_vertically(32);
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void move_center_left(environment& happy_env)
  {
    black_screen();
    move_horizontally(-32);
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void move_center_right(environment& happy_env)
  {
    black_screen();
    move_horizontally(32);
    happy_env.happy_line_env.draw_line_sets();
    cond_draw_curr(happy_env);
  }

void next_op_list(environment& happy_env)
  {
    happy_env.happy_op_env.next_list();
  }

void prev_op_list(environment& happy_env)
  {
    happy_env.happy_op_env.previous_list();
  }

void first_op_list(environment& happy_env)
  {
    happy_env.happy_op_env.first_list();
  }

void last_op_list(environment& happy_env)
  {
    happy_env.happy_op_env.last_list();
  }

void indexed_op_list(environment& happy_env)
  {
    string integer;
    if(!happy_env.is_sub_env)
    {
      integer= get_int();
    }
    else
    {
      getline(*(happy_env.script), integer);
    }
    int index= str_to_int(integer);
    happy_env.happy_op_env.indexed_list(index);
  }

void pop_last_op(environment& happy_env)
  {
    happy_env.happy_op_env.pop_back_current_list();
  }

void new_op_list(environment& happy_env)
  {
    happy_env.happy_op_env.add_new_list();
  }

void save_int_script(environment& happy_env)
  {
    if(!happy_env.is_sub_env)
    {
      happy_env.happy_op_env.save_current_op_list(get_file_name());
    }
    else
    {
      string fez;
      (*(happy_env.script)) >> fez;
      happy_env.happy_op_env.save_current_op_list(fez);
    }
  }

void use_command_line(environment& happy_env)
  {
    string raw= get_raw_command_line();
    list<string> cooked;
    cook_command_line(raw, cooked);
    happy_env.process_command_line(cooked);
  }

void load_int_script(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    string script_name;
    if(!happy_env.is_sub_env)
    {
      script_name= get_file_name();
    }
    else
    {
      (*(happy_env.script)) >> script_name;
    }
    happy_env.process_int_script(script_name);
    cond_draw_curr(happy_env);
  }

void load_int_rlswap_script(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    string script_name;
    if(!happy_env.is_sub_env)
    {
      script_name= get_file_name();
    }
    else
    {
      (*(happy_env.script)) >> script_name;
    }
    happy_env.process_int_rlswap_script(script_name);
    cond_draw_curr(happy_env);
  }

void load_command_script(environment& happy_env)
  {
    cond_undraw_curr(happy_env);
    string script_name;
    if(!happy_env.is_sub_env)
    {
      script_name= get_file_name();
    }
    else
    {
      (*(happy_env.script)) >> script_name;
    }
    happy_env.process_command_script(script_name);
    cond_draw_curr(happy_env);
  }

void load_new_template_set(environment& happy_env)
  {
    string script_name;
    if(!happy_env.is_sub_env)
    {
      script_name= get_file_name();
    }
    else
    {
      (*(happy_env.script)) >> script_name;
    }
    happy_env.happy_line_env.happy_template_env.load_new_template_set(script_name);
  }

void drop_curr_template_set(environment& happy_env)
  {
    happy_env.happy_line_env.undraw_current_line();
    happy_env.happy_line_env.happy_template_env.drop_current_template_set();
    happy_env.happy_line_env.reset_current_line();
    happy_env.happy_line_env.draw_current_line();
  }

void view_lossage(environment& happy_env)
  {
    string lost;
    for(list<SDLKey>::iterator n= happy_env.lossage.begin(); n != happy_env.lossage.end(); ++n)
    {
      lost+= sdlk_to_string(*n);
      lost+= ' ';
    }
    display_error(lost);
  }

void reset(environment& happy_env)
  {
    happy_env.mung();
  }

void leave(environment& happy_env)
  {
    happy_env.quit= !happy_env.quit; /* Just for the fun of it.  The quit operation can be multiplied.  */
  }

void save_normal(environment& happy_env)
  {
    string name= get_file_name();
    save_pic_as_png(screen, name);
  }

bool quit= false;
