/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef ENVIRONMENT_HH
#define ENVIRONMENT_HH

#include <fstream>
#include <list>
#include "SDL.h"

#include "line_env.hh"
#include "line_set.hh"
#include "op_env.hh"
#include "template_env.hh"

enum operation_ints
{
  set_and_move_int
  , set_without_moving_int
  , move_without_setting_int
  , turn_left_int
  , turn_right_int
  , trip_int
  , quad_int
  , quin_int
  , turn_to_first_int
  , turn_to_last_int
  , turn_to_indexed_int
  , meta_turn_left_int
  , meta_turn_right_int
  , meta_turn_to_first_int
  , meta_turn_to_last_int
  , meta_turn_to_indexed_int
  , inc_curr_line_red_1_int
  , inc_curr_line_red_4_int
  , inc_curr_line_red_16_int
  , inc_curr_line_red_64_int
  , dec_curr_line_red_1_int
  , dec_curr_line_red_4_int
  , dec_curr_line_red_16_int
  , dec_curr_line_red_64_int
  , zero_curr_line_red_int
  , max_curr_line_red_int
  , inc_curr_line_green_1_int
  , inc_curr_line_green_4_int
  , inc_curr_line_green_16_int
  , inc_curr_line_green_64_int
  , dec_curr_line_green_1_int
  , dec_curr_line_green_4_int
  , dec_curr_line_green_16_int
  , dec_curr_line_green_64_int
  , zero_curr_line_green_int
  , max_curr_line_green_int
  , inc_curr_line_blue_1_int
  , inc_curr_line_blue_4_int
  , inc_curr_line_blue_16_int
  , inc_curr_line_blue_64_int
  , dec_curr_line_blue_1_int
  , dec_curr_line_blue_4_int
  , dec_curr_line_blue_16_int
  , dec_curr_line_blue_64_int
  , zero_curr_line_blue_int
  , max_curr_line_blue_int
//   , inc_zoom_int
//   , dec_zoom_int
//   , move_center_up_int
//   , move_center_down_int
//   , move_center_left_int
//   , move_center_right_int
//   , save_int_script_int
//   , use_command_line_int
//   , load_int_script_int
//   , load_command_script_int
//   , load_new_template_set_int
//   , drop_curr_template_set_int
//   , drop_template_set_int
  , num_operation_ints
};

struct environment
{
  environment();

  void process_sdlk(SDLKey command);
  void process_int(int command);
  void process_command_line(std::list<std::string> command);
  void process_int_script(std::string file_name) throw(error);
  void process_int_rlswap_script(std::string file_name) throw(error);
  void process_command_script(std::string file_name);

  void interpret_int(int command);
  void interpret_sdlk(SDLKey command);
  int convert_sdlk_to_int(SDLKey command);
  bool multiplier_op(int op);
  
  void mung();
  void init(std::list<std::string>& template_file_names);

  line_env happy_line_env;
  op_env happy_op_env;

  std::list<unsigned int> next_op_multiplier;
  std::list<SDLKey> lossage;
  bool swap_r_l;
  bool is_sub_env;
  std::ifstream* script;

  bool quit;

private:
  point starting_point;
};

extern environment current_env;

/* These would be members if I hadn't heard such bad things about pointers to member
   functions. */

void turn_left(environment& happy_env);
void turn_right(environment& happy_env);
void turn_to_first(environment& happy_env);
void turn_to_last(environment& happy_env);
void turn_to_indexed(environment& happy_env);
void meta_turn_left(environment& happy_env);
void meta_turn_right(environment& happy_env);
void meta_turn_to_first(environment& happy_env);
void meta_turn_to_last(environment& happy_env);
void meta_turn_to_indexed(environment& happy_env);

void inc_curr_line_red_1(environment& happy_env);
void inc_curr_line_red_4(environment& happy_env);
void inc_curr_line_red_16(environment& happy_env);
void inc_curr_line_red_64(environment& happy_env);
void dec_curr_line_red_1(environment& happy_env);
void dec_curr_line_red_4(environment& happy_env);
void dec_curr_line_red_16(environment& happy_env);
void dec_curr_line_red_64(environment& happy_env);
void zero_curr_line_red(environment& happy_env);
void max_curr_line_red(environment& happy_env);
void inc_curr_line_green_1(environment& happy_env);
void inc_curr_line_green_4(environment& happy_env);
void inc_curr_line_green_16(environment& happy_env);
void inc_curr_line_green_64(environment& happy_env);
void dec_curr_line_green_1(environment& happy_env);
void dec_curr_line_green_4(environment& happy_env);
void dec_curr_line_green_16(environment& happy_env);
void dec_curr_line_green_64(environment& happy_env);
void zero_curr_line_green(environment& happy_env);
void max_curr_line_green(environment& happy_env);
void inc_curr_line_blue_1(environment& happy_env);
void inc_curr_line_blue_4(environment& happy_env);
void inc_curr_line_blue_16(environment& happy_env);
void inc_curr_line_blue_64(environment& happy_env);
void dec_curr_line_blue_1(environment& happy_env);
void dec_curr_line_blue_4(environment& happy_env);
void dec_curr_line_blue_16(environment& happy_env);
void dec_curr_line_blue_64(environment& happy_env);
void zero_curr_line_blue(environment& happy_env);
void max_curr_line_blue(environment& happy_env);

void set_and_move(environment& happy_env);
void set_without_moving(environment& happy_env);
void move_without_setting(environment& happy_env);

void trip_next(environment& happy_env);
void quad_next(environment& happy_env);
void quin_next(environment& happy_env);
void multiply_next(environment& happy_env, unsigned int factor);

void inc_zoom(environment& happy_env);
void dec_zoom(environment& happy_env);
void move_center_up(environment& happy_env);
void move_center_down(environment& happy_env);
void move_center_left(environment& happy_env);
void move_center_right(environment& happy_env);

void next_op_list(environment& happy_env);
void prev_op_list(environment& happy_env);
void first_op_list(environment& happy_env);
void last_op_list(environment& happy_env);
void indexed_op_list(environment& happy_env);
void pop_last_op(environment& happy_env);
void new_op_list(environment& happy_env);
void save_int_script(environment& happy_env);

void use_command_line(environment& happy_env);
void load_int_script(environment& happy_env);
void load_int_rlswap_script(environment& happy_env);
void load_command_script(environment& happy_env);
void load_new_template_set(environment& happy_env);
void drop_curr_template_set(environment& happy_env);

void view_lossage(environment& happy_env);

void reset(environment& happy_env);

void leave(environment& happy_env);

void save_normal(environment& happy_env);

extern bool quit;

#endif
