/*
  Copyright Eric Reese, 2005
  This code is distributed under the GNU GPL, version 2 or (at your option) any later
  version as published by the Free Software Foundation.  This comes with no warranty
  whatsoever, without even the implied warranty of merchantability or fitness for a
  particular purpose.  It might not even work.  See the GNU GPL for more details.
  http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef HAPPY_LIST_HH
#define HAPPY_LIST_HH

#include <deque>

#include "error.hh"
#include "str_tools.hh"

template<class LIST_EL_T, bool is_const>
  struct circle_iterator_list_el_t;
      
template<class LIST_EL_T>
  struct circle_iterator_list_el_t<LIST_EL_T, true>
  {
    typedef LIST_EL_T const type;
  };

template<class LIST_EL_T>
  struct circle_iterator_list_el_t<LIST_EL_T, false>
  {
    typedef LIST_EL_T type;
  };

template<class LIST_EL_T, bool is_const, bool is_reverse>
  struct circle_iterator_deque_iterator_t;
      
template<class LIST_EL_T>
  struct circle_iterator_deque_iterator_t<LIST_EL_T, true, true>
  {
    typedef typename std::deque<LIST_EL_T>::const_reverse_iterator type;
  };

template<class LIST_EL_T>
  struct circle_iterator_deque_iterator_t<LIST_EL_T, false, true>
  {
    typedef typename std::deque<LIST_EL_T>::reverse_iterator type;
  };

template<class LIST_EL_T>
  struct circle_iterator_deque_iterator_t<LIST_EL_T, true, false>
  {
    typedef typename std::deque<LIST_EL_T>::const_iterator type;
  };

template<class LIST_EL_T>
  struct circle_iterator_deque_iterator_t<LIST_EL_T, false, false>
  {
    typedef typename std::deque<LIST_EL_T>::iterator type;
  };

template<class LIST_EL_T, bool is_const, bool is_reverse>
  struct circle_iterator_template
  {
    circle_iterator_template();
    circle_iterator_template(
      typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_beg
      , typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_end
      , typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_cur);
    circle_iterator_template& operator++(); /* prefix */
    circle_iterator_template operator++(int); /* postfix, int is unused */
    circle_iterator_template& operator--();
    circle_iterator_template operator--(int);
    typename circle_iterator_list_el_t<LIST_EL_T, is_const>::type& operator*();
    bool at_end();
      
  private:
    typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type pos;
    typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type fake_beg;
    typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type cont_beg;
    typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type cont_end;
    int distance_traveled;
  };

template<class LIST_EL_T, bool is_const, bool is_reverse>
  circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template()
    :pos()
     , fake_beg()
     , distance_traveled(0)
  {}

template<class LIST_EL_T, bool is_const, bool is_reverse>
  circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template(
    typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_beg
    , typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_end
    , typename circle_iterator_deque_iterator_t<LIST_EL_T, is_const, is_reverse>::type c_cur)
    :pos(c_cur)
     , fake_beg(c_cur)
     , cont_beg(c_beg)
     , cont_end(c_end)
  {}

template<class LIST_EL_T, bool is_const, bool is_reverse>
  typename circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template& circle_iterator_template<LIST_EL_T, is_const, is_reverse>::operator++() /* prefix */
  {
    ++pos;
    if(pos == cont_end)
    {
      pos= cont_beg;
    }
    ++distance_traveled;
    return *this;
  }
      
template<class LIST_EL_T, bool is_const, bool is_reverse>
  typename circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template circle_iterator_template<LIST_EL_T, is_const, is_reverse>::operator++(int) /* postfix, int is unused */
  {
    typename circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template ret= *this;
    ++pos;
    if(pos == cont_end)
    {
      pos= cont_beg;
    }
    ++distance_traveled;
    return ret;
  }
      
template<class LIST_EL_T, bool is_const, bool is_reverse>
  typename circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template& circle_iterator_template<LIST_EL_T, is_const, is_reverse>::operator--()
  {
    if(pos == cont_beg)
    {
      pos= cont_end;
    }
    --pos;
    --distance_traveled;
    return *this;
  }
      
template<class LIST_EL_T, bool is_const, bool is_reverse>
  typename circle_iterator_template<LIST_EL_T, is_const, is_reverse>::circle_iterator_template circle_iterator_template<LIST_EL_T, is_const, is_reverse>::operator--(int)
  {
    circle_iterator_template ret= *this;
    if(pos == cont_beg)
    {
      pos= cont_end;
    }
    --pos;
    --distance_traveled;
    return ret;
  }
      
template<class LIST_EL_T, bool is_const, bool is_reverse>
  typename circle_iterator_list_el_t<LIST_EL_T, is_const>::type& circle_iterator_template<LIST_EL_T, is_const, is_reverse>::operator*()
  {
    return *pos;
  }

template<class LIST_EL_T, bool is_const, bool is_reverse>
  bool circle_iterator_template<LIST_EL_T, is_const, is_reverse>::at_end()
  {
    if(distance_traveled == 0)
    {
      return false;
    }
    return pos == fake_beg;
  }

template<class LIST_EL_T> struct happy_list
  {
    typedef typename std::deque<LIST_EL_T>::iterator normal_iterator;
    typedef typename std::deque<LIST_EL_T>::const_iterator const_normal_iterator;
    typedef typename std::deque<LIST_EL_T>::reverse_iterator reverse_normal_iterator;
    typedef typename std::deque<LIST_EL_T>::const_reverse_iterator const_reverse_normal_iterator;

    happy_list()
      :current_el(all_els.begin())
    {}

    happy_list(happy_list const& foo)
      :all_els(foo.all_els)
       , current_el_index(foo.current_el_index)
       , current_el(all_els.begin() + current_el_index)
    {}

    LIST_EL_T& get_current_el();
    LIST_EL_T& get_next_el();
    LIST_EL_T& get_previous_el();
    LIST_EL_T& get_indexed_el(int index) throw(error);
    LIST_EL_T& get_first_el();
    LIST_EL_T& get_last_el();
    bool select_next_el();
    bool select_previous_el();
    void select_indexed_el(int index) throw(error);
    void select_first_el();
    void select_last_el();
    void remove_current_el();
    void remove_next_el();
    void remove_previous_el();
    void remove_indexed_el(int index) throw(error);
    void remove_first_el();
    void remove_last_el();
    void resize(size_t new_size);
    void incsize();
    void push_back(LIST_EL_T& fax);
    void push_back(normal_iterator beg, normal_iterator end);
    void mung();
    void insert(normal_iterator beg, normal_iterator end);
    void insert(LIST_EL_T& src);
    bool empty();
    size_t size();

    normal_iterator begin();
    normal_iterator end();
    const_normal_iterator begin() const;
    const_normal_iterator end() const;
    reverse_normal_iterator rbegin();
    reverse_normal_iterator rend();
    const_reverse_normal_iterator rbegin() const;
    const_reverse_normal_iterator rend() const;

    typedef circle_iterator_template<LIST_EL_T, false, false> circle_iterator;
    typedef circle_iterator_template<LIST_EL_T, true, false> const_circle_iterator;
    typedef circle_iterator_template<LIST_EL_T, false, true> reverse_circle_iterator;
    typedef circle_iterator_template<LIST_EL_T, true, true> const_reverse_circle_iterator;
    circle_iterator cbegin();
    const_circle_iterator cbegin() const;
    reverse_circle_iterator rcbegin();
    const_reverse_circle_iterator rcbegin() const;

  private:
    std::deque<LIST_EL_T> all_els;
    size_t current_el_index;
    typename std::deque<LIST_EL_T>::iterator current_el;
  };

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_current_el()
  {
    return *current_el;
  }

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_next_el()
  {
    if(current_el_index == all_els.size() - 1)
    {
      return all_els.front();
    }
    else
    {
      return current_el[1]; /* Should work? */
    }
  }

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_previous_el()
  {
    if(current_el_index == 0)
    {
      return all_els.back();
    }
    else
    {
      return current_el[-1]; /* I wonder.... */
    }
  }

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_indexed_el(int index) throw(error)
  {
    return all_els[normalize_index(index, all_els.size())];
  }

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_first_el()
  {
    return all_els.front();
  }

template<class LIST_EL_T>
LIST_EL_T& happy_list<LIST_EL_T>::get_last_el()
  {
    return all_els.back();
  }

template<class LIST_EL_T>
bool happy_list<LIST_EL_T>::select_next_el()
  {
    ++current_el;
    ++current_el_index;
    if(current_el == all_els.end())
    {
      current_el= all_els.begin();
      current_el_index= 0;
      return true;
    }
    return false;
  }
    
template<class LIST_EL_T>
bool happy_list<LIST_EL_T>::select_previous_el()
  {
    bool wrapped= false;
    if(current_el == all_els.begin())
    {
      current_el= all_els.end();
      current_el_index= all_els.size();
      wrapped= true;
    }
    --current_el;
    --current_el_index;
    return wrapped;
  }
    
template<class LIST_EL_T>
void happy_list<LIST_EL_T>::select_indexed_el(int index) throw(error)
  {
    current_el= all_els.begin() + normalize_index(index, all_els.size());
    current_el_index= index;
  }
    
template<class LIST_EL_T>
void happy_list<LIST_EL_T>::select_first_el()
  {
    current_el= all_els.begin();
    current_el_index= 0;
  }
    
template<class LIST_EL_T>
void happy_list<LIST_EL_T>::select_last_el()
  {
    current_el= all_els.end() - 1;
    current_el_index= all_els.size() - 1;
  }
    
template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_current_el()
  {
    all_els.erase(current_el);
    if(current_el_index == all_els.size())
    {
      current_el_index= 0;
    }
    current_el= all_els.begin() + current_el_index;
  }
    
template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_next_el()
  {
    if(current_el_index == all_els.size() - 1)
    {
      all_els.erase(all_els.begin());
      --current_el_index;
      current_el= all_els.begin() + current_el_index;
    }
    else
    {
      all_els.erase(current_el + 1);
    }
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_previous_el()
  {
    if(current_el_index == 0)
    {
      all_els.erase(all_els.end() - 1);
    }
    else
    {
      current_el= all_els.erase(current_el - 1);
      --current_el_index;
    }
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_indexed_el(int index) throw(error)
  {
    size_t nindex= normalize_index(index, all_els.size());
    all_els.erase(all_els.begin() + nindex);
    if(current_el_index >= nindex)
    {
      if(current_el_index == all_els.size())
      {
        current_el_index= 0;
      }
      current_el= all_els.begin() + current_el_index;
    }
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_first_el()
  {
    all_els.erase(all_els.begin());
    if(current_el_index == all_els.size())
    {
      current_el_index= 0;
    }
    current_el= all_els.begin() + current_el_index;
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::remove_last_el()
  {
    all_els.erase(all_els.end() - 1);
    if(current_el_index == all_els.size())
    {
      current_el_index= 0;
    }
    current_el= all_els.begin() + current_el_index;
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::resize(size_t new_size)
  {
    all_els.resize(new_size);
    if(current_el_index >= all_els.size())
    {
      current_el_index= 0;
    }
    current_el= all_els.begin() + current_el_index;
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::incsize()
  {
    resize(size() + 1);
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::push_back(LIST_EL_T& fax)
  {
    all_els.push_back(fax);
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::push_back(normal_iterator beg, normal_iterator end)
  {
    all_els.insert(all_els.end(), beg, end);
    current_el= all_els.begin() + current_el_index;
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::mung()
  {
    all_els.clear();
    current_el= all_els.end();
    current_el_index= all_els.size();
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::insert(normal_iterator beg, normal_iterator end)
  {
    size_t prev_size= size();
    all_els.insert(current_el, beg, end);
    current_el_index+= size() - prev_size;
    current_el= all_els.begin() + current_el_index;
  }

template<class LIST_EL_T>
void happy_list<LIST_EL_T>::insert(LIST_EL_T& src)
  {
    current_el= all_els.insert(current_el, src);
    ++current_el_index;
  }

template<class LIST_EL_T>
bool happy_list<LIST_EL_T>::empty()
  {
    return all_els.empty();
  }

template<class LIST_EL_T>
size_t happy_list<LIST_EL_T>::size()
  {
    return all_els.size();
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::normal_iterator happy_list<LIST_EL_T>::begin()
  {
    return all_els.begin();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::normal_iterator happy_list<LIST_EL_T>::end()
  {
    return all_els.end();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_normal_iterator happy_list<LIST_EL_T>::begin() const
  {
    return all_els.begin();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_normal_iterator happy_list<LIST_EL_T>::end() const
  {
    return all_els.end();
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::reverse_normal_iterator happy_list<LIST_EL_T>::rbegin()
  {
    return all_els.rbegin();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::reverse_normal_iterator happy_list<LIST_EL_T>::rend()
  {
    return all_els.rend();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_reverse_normal_iterator happy_list<LIST_EL_T>::rbegin() const
  {
    return all_els.rbegin();
  }
    
template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_reverse_normal_iterator happy_list<LIST_EL_T>::rend() const
  {
    return all_els.rend();
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::circle_iterator happy_list<LIST_EL_T>::cbegin()
  {
    return circle_iterator(all_els.begin(), all_els.end(), current_el);
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_circle_iterator happy_list<LIST_EL_T>::cbegin() const
  {
    return const_circle_iterator(all_els.begin(), all_els.end(), current_el);
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::reverse_circle_iterator happy_list<LIST_EL_T>::rcbegin()
  {
    return reverse_circle_iterator(all_els.rbegin(), all_els.rend(), current_el);
  }

template<class LIST_EL_T>
typename happy_list<LIST_EL_T>::const_reverse_circle_iterator happy_list<LIST_EL_T>::rcbegin() const
  {
    return const_reverse_circle_iterator(all_els.rbegin(), all_els.rend(), current_el);
  }

#endif
