/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef LINE_HH
#define LINE_HH

#include <iostream>
#include <vector>
#include <utility>

#include "SDL.h"

#include "point.hh"

struct basic_line
{
  basic_line();
  basic_line(std::vector<point> pixies);
  std::vector<point> pixels;
};

struct line : basic_line
{
  line();
  line(line const& foo);
  
  bool is_valid() const;
  
  void draw() const;
  void draw_point(point p) const;
  void draw_ignore_special() const;
  /* Ignore all the special members (invisible, interpolated, hilighted, e_color)
     when drawing. */
  void draw_invisible() const; /* I know it seems silly. */
  void draw_interpolated() const;
  void draw_hilighted() const;
  void draw_as_color(Uint32 color) const;
  void draw_special(bool d_invis, bool d_interp, bool d_hilight, Uint32 d_s_c
    , Uint32 d_e_c) const;
  /* Logic says the d_invis parameter isn't needed, but my intuition says to provide
     it anyway, so that all fields that affect drawing can be "set" this way. */
  void draw_really_special(std::pair<bool, bool> d_invis, std::pair<bool, bool> d_interp
    , std::pair<bool, bool> d_hilight, std::pair<bool, Uint32> d_s_c
    , std::pair<bool, Uint32> d_e_c) const;
  /* The first member (bool) of each pair specifies whether the second member (value)
     should be used in the drawing. */
  
  void set_pixels(point start, basic_line& line_template, bool invis, bool interp
    , bool hilight, Uint32 s_c, Uint32 e_c);

  std::ostream& spew_info(std::ostream& out_file);
  std::istream& read_info(std::istream& in_file);
//   std::istream& read_unformatted_info(std::istream& in_file);

  Uint32 shift_channel(bool s_or_e, unsigned int channel, int amount);
  /* Only looks at the first two bits of channel and the first 8 bits of amount.
     Perhaps I should use chars instead, to get a type of minimal size. */

  /* I really don't know why I included such a wide array of interface functions.
     Even I wonder how many I'll actually use. */

  /* Screw it, has_been_set is the only one that should be private, accessor
     functions are just filler. */
  bool invisible;
  /* Just trust my intuition on this one. */
  bool interpolated;
  /* Undecided on whether I should provide interpolation, given the shortness of the
     lines, but it'll be easier to add if the data members are already there. */
  /* With the wavy line templates, the lines are significantly longer, so length of
     lines is not a valid argument against interpolation. */
  bool hilighted;
  /* My intuition tells me to trust my intuition. */
  Uint32 s_color, e_color;

private: /* Done solely because of the influence of example code I've seen. */
  bool has_been_set;
};

bool operator==(line rhs, line lhs);

#endif
