/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <fstream>
#include <sstream>
#include <utility>

#include "draw.hh"
#include "line_env.hh"
#include "str_tools.hh"
#include "template_io.hh"

using std::list;
using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::ostringstream;
using std::string;
using std::pair;

void line_env::init(point starting_point, list<string>& template_file_names)
  {
    happy_template_env.init(template_file_names);
    line_sets.resize(1);
    change_current_line(starting_point, happy_template_env.get_current_template()
      , false, false, false, 0xffffff, 0xffffff);
  }

void line_env::parse_top_level_command(list<string>& command)
  {
    if(match_strings("d", "draw", command.front()))
    {
      command.pop_front();
      parse_top_level_draw_command(command);
    }
    else if(match_strings("s", "sets", command.front()))
    {
      command.pop_front();
      parse_top_level_change_sets_command(command);
    }
    else if(match_strings("l", "line", command.front()))
    {
      command.pop_front();
      parse_top_level_change_line_command(command);
    }
  }

void line_env::parse_top_level_draw_command(list<string>& command)
  {
    if(match_strings("l", "line", command.front()))
    {
      command.pop_front();
      parse_draw_line_command(command);
    }
    else if(match_strings("s", "set", command.front()))
    {
      command.pop_front();
      parse_draw_set_command(command);
    }
  }

void line_env::parse_draw_set_command(list<string>& command)
  {
    bool apply_to_all= false;
    if(match_strings("a", "all", command.front()))
    {
      apply_to_all= true;
      command.pop_front();
    }
    else if(match_strings("d", "draw", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n!= line_sets.end(); ++n)
        {
          n->draw_all();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all();
      }
    }
    else if(match_strings("igg", "ignore_global_specials", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_ignore_global_specials();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_ignore_global_specials();
      }
    }
    else if(match_strings("ig", "ignore_specials", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_ignore_special();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_ignore_special();
      }
    }
    else if(match_strings("inv", "invisible", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_invisible();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_invisible();
      }
    }
    else if(match_strings("int", "interpolated", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_interpolated();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_interpolated();
      }
    }
    else if(match_strings("h", "hilighted", command.front()))
    {
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_hilighted();
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_hilighted();
      }
    }
    else if(match_strings("c", "color", command.front()))
    {
      command.pop_front();
      Uint32 color= str_to_uint32(command.front());
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_as_color(color);
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_as_color(color);
      }
    }
    else if(match_strings("s", "special", command.front()))
    {
      command.pop_front();
      bool inv, inter, hi;
      Uint32 sc, ec;
      inv= str_to_int(command.front());
      command.pop_front();
      inter= str_to_int(command.front());
      command.pop_front();
      hi= str_to_int(command.front());
      command.pop_front();
      sc= str_to_uint32(command.front());
      command.pop_front();
      ec= str_to_uint32(command.front());
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_special(inv, inter, hi, sc, ec);
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_special(inv, inter, hi, sc, ec);
      }
    }
    else if(match_strings("rs", "really_special", command.front()))
    {
      command.pop_front();
      pair<bool, bool> invis, interp, hilight;
      pair<bool, Uint32> s_c, e_c;
      invis.first= str_to_int(command.front());
      command.pop_front();
      invis.second= str_to_int(command.front());
      command.pop_front();
      interp.first= str_to_int(command.front());
      command.pop_front();
      interp.second= str_to_int(command.front());
      command.pop_front();
      hilight.first= str_to_int(command.front());
      command.pop_front();
      hilight.second= str_to_int(command.front());
      command.pop_front();
      s_c.first= str_to_int(command.front());
      command.pop_front();
      s_c.second= str_to_uint32(command.front());
      command.pop_front();
      e_c.first= str_to_int(command.front());
      command.pop_front();
      e_c.second= str_to_uint32(command.front());
      command.pop_front();
      if(apply_to_all)
      {
        for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
        {
          n->draw_all_really_special(invis, interp, hilight, s_c, e_c);
        }
      }
      else
      {
        line_sets.get_current_el().draw_all_really_special(invis, interp, hilight, s_c, e_c);
      }
    }
  }

void line_env::parse_draw_line_command(list<string>& command)
  {
    if(match_strings("d", "draw", command.front()))
    {
      current_line.draw();
      command.pop_front();
    }
    else if(match_strings("p", "point", command.front()))
    {
      command.pop_front();
      point fap;
      fap.x= str_to_int(command.front());
      command.pop_front();
      fap.y= str_to_int(command.front());
      current_line.draw_point(fap);
      command.pop_front();
    }
    else if(match_strings("ig", "ignore", command.front()))
    {
      current_line.draw_ignore_special();
      command.pop_front();
    }
    else if(match_strings("inv", "invisibile", command.front()))
    {
      current_line.draw_invisible();
      command.pop_front();
    }
    else if(match_strings("int", "interpolated", command.front()))
    {
      current_line.draw_interpolated();
      command.pop_front();
    }
    else if(match_strings("h", "hilighted", command.front()))
    {
      current_line.draw_hilighted();
      command.pop_front();
    }
    else if(match_strings("c", "color", command.front()))
    {
      command.pop_front();
      Uint32 fip;
      fip= str_to_uint32(command.front());
      current_line.draw_as_color(fip);
      command.pop_front();
    }
    else if(match_strings("s", "special", command.front()))
    {
      command.pop_front();
      bool invis, interp, hilight;
      Uint32 s, e;
      invis= str_to_int(command.front());
      command.pop_front();
      interp= str_to_int(command.front());
      command.pop_front();
      hilight= str_to_int(command.front());
      command.pop_front();
      s= str_to_uint32(command.front());
      command.pop_front();
      e= str_to_uint32(command.front());
      command.pop_front();
      current_line.draw_special(invis, interp, hilight, s, e);
    }
    else if(match_strings("r", "really_special", command.front()))
    {
      command.pop_front();
      pair<bool, bool> invis, interp, hilight;
      pair<bool, Uint32> s_c, e_c;
      invis.first= str_to_int(command.front());
      command.pop_front();
      invis.second= str_to_int(command.front());
      command.pop_front();
      interp.first= str_to_int(command.front());
      command.pop_front();
      interp.second= str_to_int(command.front());
      command.pop_front();
      hilight.first= str_to_int(command.front());
      command.pop_front();
      hilight.second= str_to_int(command.front());
      command.pop_front();
      s_c.first= str_to_int(command.front());
      command.pop_front();
      s_c.second= str_to_uint32(command.front());
      command.pop_front();
      e_c.first= str_to_int(command.front());
      command.pop_front();
      e_c.second= str_to_uint32(command.front());
      command.pop_front();
      current_line.draw_really_special(invis, interp, hilight, s_c, e_c);
    }
  }

void line_env::parse_top_level_change_sets_command(list<string>& command)
  {
    if(match_strings("c", "change", command.front()))
    {
      command.pop_front();
      parse_change_sets_command(command);
    }
    else if(match_strings("s", "select", command.front()))
    {
      command.pop_front();
      parse_select_different_set_command(command);
    }
  }

void line_env::parse_change_sets_command(list<string>& command)
  {
    if(match_strings("c", "current", command.front()))
    {
      command.pop_front();
      parse_change_curr_set_command(command);
    }
    else
    {
      parse_change_set_command(command, parse_set_choice_command(command));
    }
  }

void line_env::parse_select_different_set_command(list<string>& command)
  {
    if(match_strings("n", "next", command.front()))
    {
      command.pop_front();
      line_sets.select_next_el();
    }
    else if(match_strings("p", "previous", command.front()))
    {
      command.pop_front();
      line_sets.select_previous_el();
    }
    else if(match_strings("i", "indexed", command.front()))
    {
      command.pop_front();
      int index = str_to_int(command.front());
      command.pop_front();
      line_sets.select_indexed_el(index);
    }
    else if(match_strings("f", "first", command.front()))
    {
      command.pop_front();
      line_sets.select_first_el();
    }
    else if(match_strings("l", "last", command.front()))
    {
      command.pop_front();
      line_sets.select_last_el();
    }
  }

void line_env::parse_change_curr_set_command(list<string>& command)
  {
    parse_change_set_command(command, line_sets.get_current_el());
  }

void line_env::parse_change_set_command(list<string>& command, line_set& victim) throw(error)
  {
    if(match_strings("a", "access", command.front()))
    {
      command.pop_front();
      if(match_strings("b", "back", command.front()))
      {
        command.pop_front();
        parse_change_line_command(command, victim.access_back_line());
      }
      else if(match_strings("f", "front", command.front()))
      {
        command.pop_front();
        parse_change_line_command(command, victim.access_front_line());
      }
      else if(match_strings("i", "index", command.front()))
      {
        command.pop_front();
        int index;
        index= str_to_int(command.front());
        command.pop_front();
        parse_change_line_command(command, victim.access_by_index(index));
      }
      else if(match_strings("p", "point", command.front()))
      {
        command.pop_front();
        point p;
        p.x= str_to_int(command.front());
        command.pop_front();
        p.y= str_to_int(command.front());
        command.pop_front();
        parse_change_line_command(command, victim. access_by_point(p));
      }
    }
    else if(match_strings("pu", "push", command.front()))
    {
      command.pop_front();
      if(match_strings("b", "back", command.front()))
      {
        command.pop_front();
        victim.push_back_line(*(parse_line_choice_command(command)));
      }
      else if(match_strings("f", "front", command.front()))
      {
        command.pop_front();
        victim.push_front_line(*(parse_line_choice_command(command)));
      }
      else if(match_strings("i", "index", command.front()))
      {
        command.pop_front();
        int index;
        index= str_to_int(command.front());
        command.pop_front();
        victim.push_before_index(index, *(parse_line_choice_command(command)));
      }
      else if(match_strings("p", "point", command.front()))
      {
        command.pop_front();
        point p;
        p.x= str_to_int(command.front());
        command.pop_front();
        p.y= str_to_int(command.front());
        command.pop_front();
        victim.push_at_point(p, *(parse_line_choice_command(command)));
      }
    }
    else if(match_strings("po", "pop", command.front()))
    {
      command.pop_front();
      if(match_strings("b", "back", command.front()))
      {
        command.pop_front();
        victim.pop_back_line();
      }
      else if(match_strings("f", "front", command.front()))
      {
        command.pop_front();
        victim.pop_front_line();
      }
      else if(match_strings("i", "index", command.front()))
      {
        command.pop_front();
        int index;
        index= str_to_int(command.front());
        command.pop_front();
        victim.pop_by_index(index);
      }
      else if(match_strings("p", "point", command.front()))
      {
        command.pop_front();
        point p;
        p.x= str_to_int(command.front());
        command.pop_front();
        p.y= str_to_int(command.front());
        command.pop_front();
        victim.pop_by_point(p);
      }
    }
    else if(match_strings("a", "add", command.front()))
    {
      command.pop_front();
      line_set& food= parse_set_choice_command(command);
      if(&food == &victim)
      {
        throw error("Blatantly refusing to add a set to itself.");
      }
      victim.add_set(food);
    }
    else if(match_strings("m", "mung", command.front()))
    {
      command.pop_front();
      victim.mung();
    }
    else if(match_strings("si", "size", command.front()))
    {
      command.pop_front();
      string s;
      ostringstream o(s);
      o << victim.size();
      /* I plan to make the environment command parser catch exceptions and display
         their strings, so this is the simplest way to return the size to the user. */
      throw error(o.str());
    }
    else if(match_strings("sa", "save", command.front()))
    {
      command.pop_front();
      ofstream out_file(command.front().c_str());
      if(!victim.save_set_to_file(out_file))
      {
        string ivn= "Invalid file name:  ";
        throw error(ivn + command.front());
      }
      command.pop_front();
    }
    else if(match_strings("l", "load", command.front()))
    {
      command.pop_front();
      ifstream in_file(command.front().c_str());
      try
      {
        if(!victim.load_set_from_file(in_file))
        {
          /* Gets caught by the catch below, so the file name doesn't need to be
             added here. */
          throw error("Invalid file name.");
        }
      }
      catch(error e)
      { /* Catch the error to add the file name to it. */
        string err= "Error saving to file:  ";
        err+= command.front();
        err+= ":  ";
        throw error(err + e.cause);
      }
      command.pop_front();
    }
    else if(match_strings("a", "all", command.front()))
    {
      command.pop_front();
      if(match_strings("inv", "invisible", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.all_invis);
        command.pop_front();
      }
      else if(match_strings("int", "interpolated", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.all_interp);
        command.pop_front();
      }
      else if(match_strings("h", "hilighted", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.all_hilight);
        command.pop_front();
      }
      else if(match_strings("s", "s_color", command.front()))
      {
        command.pop_front();
        victim.all_s_color= str_to_uint32(command.front());
        command.pop_front();
      }
      else if(match_strings("e", "e_color", command.front()))
      {
        command.pop_front();
        victim.all_e_color= str_to_uint32(command.front());
        command.pop_front();
      }
    }
    else if(match_strings("u", "use_all", command.front()))
    {
      command.pop_front();
      if(match_strings("inv", "invisible", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_invis);
        command.pop_front();
      }
      else if(match_strings("int", "interpolated", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_interp);
        command.pop_front();
      }
      else if(match_strings("h", "hilighted", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_hilight);
        command.pop_front();
      }
      else if(match_strings("s", "s_color", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_s_color);
        command.pop_front();
      }
      else if(match_strings("e", "e_color", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_e_color);
        command.pop_front();
      }
      else if(match_strings("a", "alls", command.front()))
      {
        command.pop_front();
        operate_on_bool(command.front(), victim.use_all_alls);
        command.pop_front();
      }
    }
  }

void line_env::parse_top_level_change_line_command(list<string>& command)
  {
    if(match_strings("c", "curr", command.front()))
    {
      command.pop_front();
      parse_change_curr_line_command(command);
    }
    else
    {
      parse_change_line_command(command, parse_line_choice_command(command));
    }
  }

void line_env::parse_change_curr_line_command(list<string>& command)
  {
    parse_change_line_command(command, &current_line);
  }

void line_env::parse_change_line_command(list<string>& command, line* flort)
  {
    if(match_strings("se", "set_pixels", command.front()))
    {
      command.pop_front();
      point p;
      if(match_strings("s", "start", command.front()))
      {
        command.pop_front();
        p= flort->pixels.front();
      }
      else
      {
        p.x= str_to_int(command.front());
        command.pop_front();
        p.y= str_to_int(command.front());
        command.pop_front();
      }
      basic_line& t= happy_template_env.get_current_template();
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        t= happy_template_env.get_current_template();
      }
      else if(match_strings("l", "line", command.front()))
      {
        command.pop_front();
        convert_line_to_template(*flort, t);
      }
      else
      {
        t= happy_template_env.parse_template_choice_command(command);
      }
      bool invis;
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        invis= flort->invisible;
      }
      else
      {
        operate_on_bool(command.front(), invis);
        command.pop_front();
      }
      bool interp;
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        invis= flort->interpolated;
      }
      else
      {
        operate_on_bool(command.front(), interp);
        command.pop_front();
      }
      bool hilight;
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        invis= flort->hilighted;
      }
      else
      {
        operate_on_bool(command.front(), hilight);
        command.pop_front();
      }
      Uint32 s_c;
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        s_c= flort->s_color;
      }
      else
      {
        s_c= str_to_uint32(command.front());
        command.pop_front();
      }
      Uint32 e_c;
      if(match_strings("c", "current", command.front()))
      {
        command.pop_front();
        e_c= flort->e_color;
      }
      else
      {
        e_c= str_to_uint32(command.front());
        command.pop_front();
      }
      flort->set_pixels(p, t, invis, interp, hilight, s_c, e_c);
    }
    else if(match_strings("sp", "spew_info", command.front()))
    {
      command.pop_front();
      ofstream out_file(command.front().c_str());
      if(!flort->spew_info(out_file))
      {
        throw error("Invalid save file:  " + command.front());
      }
      command.pop_front();
    }
    else if(match_strings("r", "read_info", command.front()))
    {
      command.pop_front();
      ifstream in_file(command.front().c_str());
      if(!flort->read_info(in_file))
      {
        throw error("Invalid load file:  " + command.front());
      }
      command.pop_front();
    }
    else if(match_strings("sh", "shift_channel", command.front()))
    {
      command.pop_front();
      bool s_or_e= str_to_bool(command.front());
      command.pop_front();
      unsigned int channel= str_to_uint32(command.front());
      command.pop_front();
      int amount= str_to_int(command.front());
      command.pop_front();
      flort->shift_channel(s_or_e, channel, amount);
    }
    else if(match_strings("inv", "invisible", command.front()))
    {
      command.pop_front();
      operate_on_bool(command.front(), flort->invisible);
      command.pop_front();
    }
    else if(match_strings("int", "interpolated", command.front()))
    {
      command.pop_front();
      operate_on_bool(command.front(), flort->interpolated);
      command.pop_front();
    }
    else if(match_strings("hi", "hilighted", command.front()))
    {
      command.pop_front();
      operate_on_bool(command.front(), flort->hilighted);
      command.pop_front();
    }
    else if(match_strings("sc", "start_color", command.front()))
    {
      command.pop_front();
      flort->s_color= str_to_uint32(command.front());
      command.pop_front();
    }
    else if(match_strings("ec", "end_color", command.front()))
    {
      command.pop_front();
      flort->e_color= str_to_uint32(command.front());
      command.pop_front();
    }
  }

line_set& line_env::parse_set_choice_command(list<string>& command) throw(error)
  {
    if(match_strings("n", "next", command.front()))
    {
      command.pop_front();
      return line_sets.get_next_el();
    }
    if(match_strings("p", "previous", command.front()))
    {
      command.pop_front();
      return line_sets.get_previous_el();
    }
    if(match_strings("i", "indexed", command.front()))
    {
      command.pop_front();
      int index= str_to_int(command.front());
      command.pop_front();
      return line_sets.get_indexed_el(index);
    }
    if(match_strings("f", "first", command.front()))
    {
      command.pop_front();
      return line_sets.get_first_el();
    }
    if(match_strings("l", "last", command.front()))
    {
      command.pop_front();
      return line_sets.get_last_el();
    }
    throw error("Invalid set choice:  " + command.front());
  }

line* line_env::parse_line_choice_command(list<string>& command) throw(error)
  {
    if(match_strings("cc", "c_current", command.front()))
    {
      command.pop_front();
      return &current_line;
    }
    line_set& set_choice= line_sets.get_current_el();
    if(match_strings("c", "current", command.front()))
    {
      command.pop_front();
      set_choice= line_sets.get_current_el();
    }
    else
    {
      set_choice= parse_set_choice_command(command);
    }
    if(match_strings("c", "current", command.front()))
    {
      command.pop_front();
      return &current_line;
    }
    if(match_strings("b", "back", command.front()))
    {
      command.pop_front();
      return set_choice.access_back_line();
    }
    if(match_strings("f", "front", command.front()))
    {
      command.pop_front();
      return set_choice.access_front_line();
    }
    if(match_strings("i", "index", command.front()))
    {
      command.pop_front();
      int index= str_to_int(command.front());
      command.pop_front();
      return set_choice.access_by_index(index);
    }
    if(match_strings("p", "point", command.front()))
    {
      command.pop_front();
      point p;
      p.x= str_to_int(command.front());
      command.pop_front();
      p.y= str_to_int(command.front());
      command.pop_front();
      return set_choice.access_by_point(p);
    }
    throw error("Invalid choice:  " + command.front());
  }

void line_env::draw_line_sets()
  {
    if(line_sets.empty())
    {
      throw error("No line sets exist!");
    }
    for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
    {
      if(!n->empty())
      {
        n->draw_all();
      }
    }
  }

void line_env::draw_last_line()
  {
    line_sets.get_current_el().access_back_line()->draw();
  }

void line_env::draw_current_set()
  {
    line_sets.get_current_el().draw_all();
  }

void line_env::draw_current_line()
  {
    current_line.draw_hilighted();
  }

void line_env::undraw_current_line()
  {
    current_line.draw_as_color(get_background_color());
    for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
    {
      if(!n->empty())
      {
        n->redraw_points(current_line.pixels.begin(), current_line.pixels.end());
      }
    }
  }

void line_env::next_set()
  {
    line_sets.select_next_el();
  }

void line_env::previous_set()
  {
    line_sets.select_previous_el();
  }

void line_env::first_set()
  {
    line_sets.select_first_el();
  }

void line_env::last_set()
  {
    line_sets.select_last_el();
  }

void line_env::indexed_set(int index)
  {
    line_sets.select_indexed_el(index);
  }

void line_env::add_curr_line_to_curr_set()
  {
    line_sets.get_current_el().push_back_line(current_line);
  }

void line_env::save_curr_set(std::string& fn) throw(error)
  {
    ofstream out_file(fn.c_str());
    if(!line_sets.get_current_el().save_set_to_file(out_file))
    {
      string s= "Invalid save file:  ";
      throw error(s + fn);
    }
  }

void line_env::load_curr_set(std::string& fn) throw(error)
  {
    ifstream in_file(fn.c_str());
    try
    {
    if(!line_sets.get_current_el().load_set_from_file(in_file))
    {
      string s= "Invalid load file:  ";
      throw error(s + fn);
    }
    }
    catch(error e)
    {
      string s= "Error loading ";
      throw error(s + fn + e.cause);
    }
  }

void line_env::mung_curr_set()
  {
    line_sets.get_current_el().mung();
    draw_line_sets();
  }

void line_env::mung_all()
  {
    for(happy_list<line_set>::normal_iterator n= line_sets.begin(); n != line_sets.end(); ++n)
    {
      n->mung();
    }
  }

point line_env::current_line_start()
  {
    return current_line.pixels.front();
  }

point line_env::current_line_end()
  {
    return current_line.pixels.back();
  }

void line_env::reset_current_line()
  {
    current_line.set_pixels(current_line_start()
      , happy_template_env.get_current_template(), current_line.invisible
      , current_line.interpolated, current_line.hilighted, current_line.s_color
      , current_line.e_color);
  }

void line_env::change_current_line(point start, basic_line& line_template, bool invis
    , bool interp, bool hilight, Uint32 s_c, Uint32 e_c)
  {
    current_line.set_pixels(start, line_template, invis, interp, hilight, s_c, e_c);
  }

void line_env::change_current_line(point start, basic_line& line_template)
  {
    current_line.set_pixels(start, line_template, current_line.invisible
      , current_line.interpolated, current_line.hilighted, current_line.s_color
      , current_line.e_color);
  }

void line_env::change_current_line(basic_line& line_template)
  {
    current_line.set_pixels(current_line_start(), line_template, current_line.invisible
      , current_line.interpolated, current_line.hilighted, current_line.s_color
      , current_line.e_color);
  }

void line_env::change_current_line(unsigned int prop_val, unsigned int prop)
  {
    switch(prop)
    {
      case 0:
        current_line.invisible= prop_val;
        break;
      case 1:
        current_line.interpolated= prop_val;
        break;
      case 2:
        current_line.hilighted= prop_val;
        break;
      case 3:
        current_line.s_color= (current_line.s_color & (~(255 << 16)))
          | ((prop_val & 255) << 16);
        break;
      case 4:
        current_line.s_color= (current_line.s_color & (~(255 << 8)))
          | ((prop_val & 255) << 8);
        break;
      case 5:
        current_line.s_color= (current_line.s_color & (~255))
          | (prop_val & 255);
        break;
      case 6:
        current_line.e_color= (current_line.e_color & (~(255 << 16)))
          | ((prop_val & 255) << 16);
        break;
      case 7:
        current_line.e_color= (current_line.e_color & (~(255 << 8)))
          | ((prop_val & 255) << 8);
        break;
      case 8:
        current_line.e_color= (current_line.e_color & (~255))
          | (prop_val & 255);
        break;
      case 9:
        current_line.s_color= prop_val;
        break;
      case 10:
        current_line.e_color= prop_val;
        break;
      default:
        break;
    }
  }

void line_env::shift_current_line_color(bool s_or_e, unsigned int channel, int amount)
  {
    current_line.shift_channel(s_or_e, channel, amount);
  }
