/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef LINE_ENV_HH
#define LINE_ENV_HH

#include <list>
#include <string>

#include "happy_list.hh"
#include "line_set.hh"
#include "template_env.hh"

struct line_env
{
  line_env()
  {}

  void init(point starting_point, std::list<std::string>& template_file_names);

  void parse_top_level_command(std::list<std::string>& command);

  void parse_top_level_draw_command(std::list<std::string>& command);
  void parse_draw_set_command(std::list<std::string>& command);
  void parse_draw_line_command(std::list<std::string>& command);

  void parse_top_level_change_sets_command(std::list<std::string>& command);
  void parse_change_sets_command(std::list<std::string>& command);
  void parse_select_different_set_command(std::list<std::string>& command);
  void parse_change_curr_set_command(std::list<std::string>& command);
  void parse_change_set_command(std::list<std::string>& command, line_set& victim)
    throw(error);

  void parse_top_level_change_line_command(std::list<std::string>& command);
  void parse_change_curr_line_command(std::list<std::string>& command);
  void parse_change_line_command(std::list<std::string>& command, line* flort);

  line_set& parse_set_choice_command(std::list<std::string>& command) throw(error);
  line* parse_line_choice_command(std::list<std::string>& command) throw(error);


  void draw_line_sets();
  void draw_last_line();
  void draw_current_set();
  void draw_current_line();
  void undraw_current_line();

  void next_set();
  void previous_set();
  void first_set();
  void last_set();
  void indexed_set(int index);
  void add_curr_line_to_curr_set();
  void save_curr_set(std::string& fn) throw(error);
  void load_curr_set(std::string& fn) throw(error);
  void mung_curr_set();
  void mung_all();

  point current_line_start();
  point current_line_end();
  void reset_current_line();
  void change_current_line(point start, basic_line& line_template, bool invis
    , bool interp, bool hilight, Uint32 s_c, Uint32 e_c);
  void change_current_line(point start, basic_line& line_template);
  void change_current_line(basic_line& line_template);
  void change_current_line(unsigned int prop_val, unsigned int prop);
  void shift_current_line_color(bool s_or_e, unsigned int channel, int amount);

  /* I dislike doing this, but it's the only way to provide for choosing a template
     when changing a line. */
  template_env happy_template_env;

private:
  happy_list<line_set> line_sets;
  line current_line;
};

#endif
