/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <algorithm>
#include <iostream>
#include <string>

#include "error.hh"
#include "line_set.hh"

using std::vector;
using std::pair;
using std::map;
using std::find;
using std::istream;
using std::ostream;
using std::cerr;
using std::endl;
using std::string;

line_set::~line_set()
  {}

line* line_set::access_back_line()
  {
    return actual_lines.back();
  }

line* line_set::access_front_line()
  {
    return actual_lines.front();
  }

line* line_set::access_by_index(int index)
  { /* Dirt slow, I know. */
    if(index < 0)
    {
      index+= actual_lines.size();
    }
    if(index < 0 || static_cast<unsigned int>(index) >= actual_lines.size())
    {
      throw error("Out of range index.");
    }
    return actual_lines[index];
  }

line* line_set::access_by_point(point pixel_of_line) throw(error)
  {
    map<point, line*>::iterator l= lines_by_point.find(pixel_of_line);
    if(l != lines_by_point.end())
    {
      return l->second;
    }
    throw error("You missed.");
  }

void line_set::push_back_line(line& foo)
  {
    line* line_to_add= new line(foo);
    actual_lines.push_back(line_to_add);
    add_points_to_map_with_line(foo.pixels.begin(), foo.pixels.end()
      , actual_lines.back());
  }

void line_set::push_front_line(line& foo)
  {
    line* line_to_add= new line(foo);
    actual_lines.push_front(line_to_add);
    add_points_to_map_with_line(foo.pixels.begin(), foo.pixels.end()
      , actual_lines.front());
  }

void line_set::push_before_index(int index, line& foo)
  {
    if(index < 0)
    {
      index+= actual_lines.size();
    }
    if(index < 0 || static_cast<unsigned int>(index) >= actual_lines.size())
    {
      throw error("Out of range index.");
    }
    line_container_t::iterator iter= actual_lines.begin() + index;
    line* line_to_add= new line(foo);
    actual_lines.insert(iter, line_to_add);
    add_points_to_map_with_line(foo.pixels.begin(), foo.pixels.end(), line_to_add);
  }

void line_set::push_at_point(point pos, line& foo) throw(error)
  {
    map<point, line*>::iterator l= lines_by_point.find(pos);
    if(l == lines_by_point.end())
    {
      throw error("You missed.");
    }
    line_container_t::iterator rl= find(actual_lines.begin(), actual_lines.end()
      , l->second);
    line* line_to_add= new line(foo);
    actual_lines.insert(rl, line_to_add);
    add_points_to_map_with_line(foo.pixels.begin(), foo.pixels.end(), line_to_add);
  }

line line_set::pop_back_line()
  {
    line ret= *(actual_lines.back());
    delete actual_lines.back();
    actual_lines.pop_back();
    remove_points_from_map(ret.pixels.begin(), ret.pixels.end());
    return ret;
  }

line line_set::pop_front_line()
  {
    line ret= *(actual_lines.front());
    delete actual_lines.front();
    actual_lines.pop_front();
    remove_points_from_map(ret.pixels.begin(), ret.pixels.end());
    return ret;
  }

line line_set::pop_by_index(int index)
  {
    if(index < 0)
    {
      index+= actual_lines.size();
    }
    if(index < 0 || static_cast<unsigned int>(index) >= actual_lines.size())
    {
      throw error("Out of range index.");
    }
    line_container_t::iterator iter= actual_lines.begin() + index;
    line ret= *(*iter);
    delete (*iter);
    actual_lines.erase(iter);
    remove_points_from_map(ret.pixels.begin(), ret.pixels.end());
    return ret;
  }

line line_set::pop_by_point(point pixel_of_line) throw(error)
  {
    map<point, line*>::iterator l= lines_by_point.find(pixel_of_line);
    if(l != lines_by_point.end())
    {
      line ret= *(l->second);
      line_container_t::iterator tmp= find(actual_lines.begin(), actual_lines.end()
        , l->second);
      if(tmp == actual_lines.end())
      {
        throw error("Line accessed by point is not in list.");
      }
      delete (*tmp);
      actual_lines.erase(tmp);
      remove_points_from_map(ret.pixels.begin(), ret.pixels.end());
      return ret;
    }
    throw error("You missed.");
  }

void line_set::add_set(line_set& set_to_add)
  {
    for(line_container_t::iterator src_iter= set_to_add.actual_lines.begin();
        src_iter != set_to_add.actual_lines.end(); ++src_iter)
    {
      actual_lines.push_back(*src_iter);
      add_points_to_map_with_line((*src_iter)->pixels.begin()
        , (*src_iter)->pixels.end(), *src_iter);
    }
  }

void line_set::mung()
  {
    all_invis= false;
    all_interp= false;
    all_hilight= false;
    all_s_color= 0xffffff;
    all_e_color= all_s_color;
    use_all_invis= false;
    use_all_interp= false;
    use_all_hilight= false;
    use_all_s_color= false;
    use_all_e_color= false;
    use_all_alls= false;
    lines_by_point.clear();
    for(line_container_t::iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      delete (*n);
    }
    actual_lines.clear();
    overlapped_points.clear();
  }

line_set::line_container_t::size_type line_set::size()
  {
    return actual_lines.size();
  }

bool line_set::empty() const
  {
    return actual_lines.empty();
  }

ostream& line_set::save_set_to_file(ostream& out_file)
  { /* Just shove it all out with spaces seperating each part. */
    if(out_file)
    {
      out_file << "# This is not a comment, that # symbol means nothing." << endl
        << "The tags don't mean anything either, all of them must be there," << endl
        << "and they must be in the right order.  They exist solely to make" << endl
        << "the file easier to navigate in a text editor for those that want to do so." << endl
        << "If you must have comments, put them at the beginning of the file," << endl
        << "before the [line_set] tag.  The file reading code throws away" << endl
        << "everything until it gets to the line containing that tag, and nothing else." << endl;
      out_file << "[line_set]" << endl << "[drawing modifiers]" << endl;
      out_file << all_invis << ' ' << all_interp << ' ' << all_hilight << ' '
        << all_s_color << ' ' << all_e_color << ' ' << use_all_invis << ' '
        << use_all_interp << ' ' << use_all_hilight << ' ' << use_all_s_color << ' '
        << use_all_e_color << ' ' << use_all_alls << endl;
      out_file << "[lines]" << endl;
      for(line_container_t::iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
      {
        ((*n)->spew_info(out_file));
      }
    }
    return out_file;
  }

istream& line_set::load_set_from_file(istream& in_file) throw(error)
  {
    if(in_file)
    {
      string s;
      while(getline(in_file, s))
      {
        if(s == "[line_set]")
        {
          break;
        }
      }
      getline(in_file, s);
      if(s != "[drawing modifiers]")
      {
        throw error("Missing or invalid [drawing modifiers] tag:  " + s);
      }
      if(!(in_file >> all_invis))
      {
        throw error("Invalid all_invis value.");
      }
      if(!(in_file >> all_interp))
      {
        throw error("Invalid all_interp value.");
      }
      if(!(in_file >> all_hilight))
      {
        throw error("Invalid all_hilight value.");
      }
      if(!(in_file >> all_s_color))
      {
        throw error("Invalid all_s_color value.");
      }
      if(!(in_file >> all_e_color))
      {
        throw error("Invalid all_e_color value.");
      }
      if(!(in_file >> use_all_invis))
      {
        throw error("Invalid use_all_invis value.");
      }
      if(!(in_file >> use_all_interp))
      {
        throw error("Invalid use_all_interp value.");
      }
      if(!(in_file >> use_all_hilight))
      {
        throw error("Invalid use_all_hilight value.");
      }
      if(!(in_file >> use_all_s_color))
      {
        throw error("Invalid use_all_s_color value.");
      }
      if(!(in_file >> use_all_e_color))
      {
        throw error("Invalid use_all_e_color value.");
      }
      if(!(in_file >> use_all_alls))
      {
        throw error("Invalid use_all_alls value.");
      }
      getline(in_file, s);
      if(s != "[lines]")
      {
        throw error("Missing or invalid [lines] tag:  " + s);
      }
      line current_line;
      while(current_line.read_info(in_file))
      {
        push_back_line(current_line);
      }
    }
    return in_file;
  }

// ostream& save_set_unformatted(std::ostream& out_file)
//   {
//     if(out_file)
//     {
//       char current_byte;
//       current_byte= all_invis | (all_interp << 1) | (all_hilight << 2);
//       out_file << current_byte << all_s_color << all_e_color;
//       current_byte= use_all_invis | (use_all_interp << 1) | (use_all_hilight << 2)
//         | (use_all_s_color << 3) | (use_all_e_color << 4) | (use_all_alls << 5);
//       out_file << current_byte;
//       for(line_container_t::iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
//       {
//         ((*n)->spew_unformatted(out_file));
//       }
//     }
//     return out_file;
//   }

// istream& load_set_unformatted(std::istream& in_file, int num_lines= -1)
//   {
//     if(in_file)
//     {
//       unsigned char current_byte;
//       in_file.get(current_byte);
//       all_invis= current_byte & 1;
//       all_interp= current_byte & 2;
//       all_hilight= current_byte & 4;
//       in_file.get(current_byte);
//       all_s_color= current_byte;
//       in_file.get(current_byte);
//       all_s_color+= (current_byte << 8);
//       in_file.get(current_byte);
//       all_s_color+= (current_byte << 16);
//       in_file.get(current_byte);
//       all_e_color= current_byte;
//       in_file.get(current_byte);
//       all_e_color+= (current_byte << 8);
//       in_file.get(current_byte);
//       all_e_color+= (current_byte << 16);
//       in_file.get(current_byte);
//       use_all_invis= current_byte & 1;
//       use_all_interp= current_byte & 2;
//       use_all_hilight= current_byte & 4;
//       use_all_s_color= current_byte & 8;
//       use_all_e_color= current_byte & 16;
//       use_all_alls= current_byte & 32;
//       if(num_lines == -1)
//       {
//         line temp;
//         while(temp.read_unformatted(in_file))
//         {
//
//         }
//       }
//     }
//     return in_file;
//   }

void line_set::draw_all() const
  { /* FINISHME:  Make this function take the all_* members into consideration. */
    if(empty())
    {
      throw error("This line set is empty!");
    }
//     cerr << "Drawing lines in set?" << endl;
    /* FIXME:  Invisibility checking it broken. */
//     if(!all_invis || (!use_all_invis && !use_all_alls))
//     {
//       cerr << "Drawing lines in set." << endl;
//       bool d_interp= all_interp && (use_all_interp || use_all_alls);
//       bool d_hilight= all_hilight && (use_all_hilight || use_all_alls);
      for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
      {
//         if(!use_all_interp && !use_all_hilight && !use_all_s_color
//           && !use_all_e_color && !use_all_alls)
//         { /* Use whatever info the line contains. */
          (*n)->draw();
//           continue;
//         }
//         pair<bool, bool> d_invis(false, false);
//         pair<bool, bool> d_interp(use_all_hilight || use_all_alls);
//         if(d_interp && !d_hilight)
//         {
//           n->draw_interpolated(dest);
//           continue;
//         }
//         if(d_hilight && !d_interp)
//         {
//           n->draw_hilighted(dest);
//           continue;
//         }
      }
//     }
  }

void line_set::draw_all_ignore_global_specials() const
  {
    for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      (*n)->draw();
    }
  }

void line_set::draw_all_ignore_special() const
  {
    for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      (*n)->draw_ignore_special();
    }
  }

void line_set::draw_all_invisible() const /* Silly, isn't it? */
  { /* If it's invisible, it has no effect on how stuff behind it looks. */ }

void line_set::draw_all_interpolated() const
  {
    for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      (*n)->draw_interpolated();
    }
  }

void line_set::draw_all_hilighted() const
  {
    for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      (*n)->draw_hilighted();
    }
  }

void line_set::draw_all_as_color(Uint32 color) const
  {
    if(!all_invis || !use_all_invis)
    {
      for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
      {
        (*n)->draw_as_color(color);
      }
    }
  }

void line_set::draw_all_special(bool d_invis, bool d_interp, bool d_hilight
  , Uint32 d_s_c, Uint32 d_e_c) const
  {
    if(!d_invis)
    {
      for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
      {
        (*n)->draw_special(d_invis, d_interp, d_hilight, d_s_c, d_e_c);
      }
    }
  }

void line_set::draw_all_really_special(pair<bool, bool> d_invis
    , pair<bool, bool> d_interp, pair<bool, bool> d_hilight, pair<bool, Uint32> d_s_c
    , pair<bool, Uint32> d_e_c) const
  {
    for(line_container_t::const_iterator n= actual_lines.begin(); n != actual_lines.end(); ++n)
    {
      (*n)->draw_really_special(d_invis, d_interp, d_hilight, d_s_c, d_e_c);
    }
  }
