/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <fstream>
#include <iostream>
#include <sstream>

#include "op_env.hh"

using std::string;
using std::list;
using std::endl;

void op_env::init()
  {
    op_lists.resize(1);
  }

void op_env::next_list()
  {
    op_lists.select_next_el();
  }

void op_env::previous_list()
  {
    op_lists.select_previous_el();
  }

void op_env::first_list()
  {
    op_lists.select_first_el();
  }

void op_env::last_list()
  {
    op_lists.select_last_el();
  }

void op_env::indexed_list(int index)
  {
    op_lists.select_indexed_el(index);
  }

void op_env::mung()
  {
    op_lists.mung();
    init();
  }

void op_env::add_op_to_current_list(int op)
  {
    op_lists.get_current_el().push_back(op);
  }

void op_env::pop_back_current_list()
  {
    if(!op_lists.get_current_el().empty())
      op_lists.get_current_el().pop_back();
  }

void op_env::add_new_list()
  {
    op_lists.incsize();
    op_lists.select_last_el();
  }

void op_env::save_op_list(string file_name, list<int>& src) throw(error)
  {
    std::ofstream out_file(file_name.c_str());
    if(!out_file)
    {
      throw error("Invalid file:  " + file_name);
    }
    out_file << "Any comments must be at the beginning.  Do not change the" << endl
      << "line containing the [ops] tag in any way.  If you are reading this" << endl
      << "file because there isn't an undo feature, you need to implement one." << endl;
    out_file << "[ops]" << endl;
    for(list<int>::iterator n= src.begin();
        n != src.end(); ++n)
    {
      out_file << *n << ' ';
    }
  }

void op_env::save_current_op_list(string file_name) throw(error)
  {
    save_op_list(file_name, op_lists.get_current_el());
  }

void op_env::save_all_op_lists(string& fn_prefix, string& fn_suffix) throw(error)
  {
    int i= 0;
    for(happy_list<list<int> >::normal_iterator n= op_lists.begin(); n != op_lists.end(); ++n)
    {
      string curr_fn;
      std::ostringstream effen_fn(curr_fn);
      effen_fn << fn_prefix << '_' << i << '_' << fn_suffix;
      save_op_list(effen_fn.str(), *n);
    }
  }
