/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <string>
#include <stdio.h>
#include "png.h"
#include "SDL.h"

#include "error.hh"
#include "png_save.hh"

using std::string;

#include <iostream>
using std::cerr;
using std::endl;

void save_pic_as_png(SDL_Surface* picture, string pic_name) throw(error)
  {
    FILE *out_file= fopen(pic_name.c_str(), "wb");
    if(!out_file)
    {
      throw error("out_file is invalid.  Aborting save.");
    }
    png_structp out_png_ptr= png_create_write_struct(PNG_LIBPNG_VER_STRING
      , NULL, NULL, NULL);
    if(!out_png_ptr)
    {
      throw error("Could not initialize png_struct.  Aborting save.");
    }
    png_infop info_ptr= png_create_info_struct(out_png_ptr);
    if(!info_ptr)
    {
      png_destroy_write_struct(&out_png_ptr, (png_infopp)NULL);
      throw error("Could not initialize png_info.  Aborting save.");
    }
    if(setjmp(png_jmpbuf(out_png_ptr)))
    {
      png_destroy_write_struct(&out_png_ptr, &info_ptr);
      fclose(out_file);
      throw error("Could not setjmp.  Aborting save.");
    }
    png_init_io(out_png_ptr, out_file);
    //  png_set_write_status_fn(out_png_ptr, write_row_callback);
    png_set_IHDR(out_png_ptr, info_ptr, picture->w, picture->h, 8
      , PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT
      , PNG_FILTER_TYPE_DEFAULT);
    png_write_info(out_png_ptr, info_ptr);
    png_uint_32 pw= picture->w, ph= picture->h;
    if(ph > PNG_UINT_32_MAX / png_sizeof(png_bytep))
    {
      /* Do something funky like making multiple pngs?  */
      png_destroy_write_struct(&out_png_ptr, &info_ptr);
      fclose(out_file);
      throw error("Too big, jerkwad.");
    }
    png_byte *image= static_cast<png_byte*>(malloc(ph*pw*4));
    png_bytep row_pointers[ph];
    for(unsigned int n= 0; n < ph*pw; n++)
    { /* Check for endianness problems when porting. */
      /* Apparently, in the pixel data for an SDL_Surface, bytes are stored BGRA,
         but in a png, RGBA.  We don't care about the alpha channel, so it gets set
         directly. */
/* R */ image[n*4+0]= (*(static_cast<Uint8*>(picture->pixels) + (n*4) + 2));
/* G */ image[n*4+1]= (*(static_cast<Uint8*>(picture->pixels) + (n*4) + 1));
/* B */ image[n*4+2]= (*(static_cast<Uint8*>(picture->pixels) + (n*4) + 0));
/* A */ image[n*4+3]= 255; /* Fully opaque. */
    }
    for(png_uint_32 n= 0; n < ph; n++)
    {
      row_pointers[n]= image + (n * pw * 4);
    }
    png_write_image(out_png_ptr, row_pointers);
    png_write_end(out_png_ptr, info_ptr);
    png_destroy_write_struct(&out_png_ptr, &info_ptr);
    fclose(out_file);
    cerr << "Picture saved to " << pic_name << " successfully." << endl;
  }
