/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <cctype>
#include <map>
#include <iostream>

#include "SDL.h"
#include "SDL_nfnt.h"

#include "draw.hh"
#include "sdl_command_line.hh"

using std::map;
using std::string;
using std::cerr;
using std::endl;

string sdlk_to_string(SDLKey sdlk)
  {
    static map<SDLKey, string> char_keysyms;
    static bool map_inited= false;
    if(!map_inited)
    {
      char_keysyms[SDLK_a]= 'a';
      char_keysyms[SDLK_b]= 'b';
      char_keysyms[SDLK_c]= 'c';
      char_keysyms[SDLK_d]= 'd';
      char_keysyms[SDLK_e]= 'e';
      char_keysyms[SDLK_f]= 'f';
      char_keysyms[SDLK_g]= 'g';
      char_keysyms[SDLK_h]= 'h';
      char_keysyms[SDLK_i]= 'i';
      char_keysyms[SDLK_j]= 'j';
      char_keysyms[SDLK_k]= 'k';
      char_keysyms[SDLK_l]= 'l';
      char_keysyms[SDLK_m]= 'm';
      char_keysyms[SDLK_n]= 'n';
      char_keysyms[SDLK_o]= 'o';
      char_keysyms[SDLK_p]= 'p';
      char_keysyms[SDLK_q]= 'q';
      char_keysyms[SDLK_r]= 'r';
      char_keysyms[SDLK_s]= 's';
      char_keysyms[SDLK_t]= 't';
      char_keysyms[SDLK_u]= 'u';
      char_keysyms[SDLK_v]= 'v';
      char_keysyms[SDLK_w]= 'w';
      char_keysyms[SDLK_x]= 'x';
      char_keysyms[SDLK_y]= 'y';
      char_keysyms[SDLK_z]= 'z';
      char_keysyms[SDLK_BACKQUOTE]= '`';
      char_keysyms[SDLK_1]= '1';
      char_keysyms[SDLK_2]= '2';
      char_keysyms[SDLK_3]= '3';
      char_keysyms[SDLK_4]= '4';
      char_keysyms[SDLK_5]= '5';
      char_keysyms[SDLK_6]= '6';
      char_keysyms[SDLK_7]= '7';
      char_keysyms[SDLK_8]= '8';
      char_keysyms[SDLK_9]= '9';
      char_keysyms[SDLK_0]= '0';
      char_keysyms[SDLK_MINUS]= '-';
      char_keysyms[SDLK_EQUALS]= '=';
      char_keysyms[SDLK_LEFTBRACKET]= '[';
      char_keysyms[SDLK_RIGHTBRACKET]= ']';
      char_keysyms[SDLK_BACKSLASH]= '\\';
      char_keysyms[SDLK_SEMICOLON]= ';';
      char_keysyms[SDLK_QUOTE]= '\'';
      char_keysyms[SDLK_COMMA]= ',';
      char_keysyms[SDLK_PERIOD]= '.';
      char_keysyms[SDLK_SLASH]= '/';
      char_keysyms[SDLK_SPACE]= ' ';
      char_keysyms[SDLK_KP0]= "KP0";
      char_keysyms[SDLK_KP1]= "KP1";
      char_keysyms[SDLK_KP2]= "KP2";
      char_keysyms[SDLK_KP3]= "KP3";
      char_keysyms[SDLK_KP4]= "KP4";
      char_keysyms[SDLK_KP5]= "KP5";
      char_keysyms[SDLK_KP6]= "KP6";
      char_keysyms[SDLK_KP7]= "KP7";
      char_keysyms[SDLK_KP8]= "KP8";
      char_keysyms[SDLK_KP9]= "KP9";
    }
    return char_keysyms[sdlk];
  }

string get_raw_command_line() throw(error)
  {
    static map<SDLKey, char> char_keysyms;
    static map<SDLKey, char> uppercase_keysyms;
    static NFNT* nfnt_font;
    static BITFONT* normal_font;
    static SDL_Rect line_pos;
    static bool font_inited_successfully= false;
    static bool static_vars_inited= false;
    if(!static_vars_inited)
    {
      line_pos.x= 0;
      line_pos.y= screen->h - 16;
      line_pos.w= 0;
      line_pos.h= 16;
      char_keysyms[SDLK_a]= 'a';
      char_keysyms[SDLK_b]= 'b';
      char_keysyms[SDLK_c]= 'c';
      char_keysyms[SDLK_d]= 'd';
      char_keysyms[SDLK_e]= 'e';
      char_keysyms[SDLK_f]= 'f';
      char_keysyms[SDLK_g]= 'g';
      char_keysyms[SDLK_h]= 'h';
      char_keysyms[SDLK_i]= 'i';
      char_keysyms[SDLK_j]= 'j';
      char_keysyms[SDLK_k]= 'k';
      char_keysyms[SDLK_l]= 'l';
      char_keysyms[SDLK_m]= 'm';
      char_keysyms[SDLK_n]= 'n';
      char_keysyms[SDLK_o]= 'o';
      char_keysyms[SDLK_p]= 'p';
      char_keysyms[SDLK_q]= 'q';
      char_keysyms[SDLK_r]= 'r';
      char_keysyms[SDLK_s]= 's';
      char_keysyms[SDLK_t]= 't';
      char_keysyms[SDLK_u]= 'u';
      char_keysyms[SDLK_v]= 'v';
      char_keysyms[SDLK_w]= 'w';
      char_keysyms[SDLK_x]= 'x';
      char_keysyms[SDLK_y]= 'y';
      char_keysyms[SDLK_z]= 'z';
      char_keysyms[SDLK_BACKQUOTE]= '`';
      char_keysyms[SDLK_1]= '1';
      char_keysyms[SDLK_2]= '2';
      char_keysyms[SDLK_3]= '3';
      char_keysyms[SDLK_4]= '4';
      char_keysyms[SDLK_5]= '5';
      char_keysyms[SDLK_6]= '6';
      char_keysyms[SDLK_7]= '7';
      char_keysyms[SDLK_8]= '8';
      char_keysyms[SDLK_9]= '9';
      char_keysyms[SDLK_0]= '0';
      char_keysyms[SDLK_MINUS]= '-';
      char_keysyms[SDLK_EQUALS]= '=';
      char_keysyms[SDLK_LEFTBRACKET]= '[';
      char_keysyms[SDLK_RIGHTBRACKET]= ']';
      char_keysyms[SDLK_BACKSLASH]= '\\';
      char_keysyms[SDLK_SEMICOLON]= ';';
      char_keysyms[SDLK_QUOTE]= '\'';
      char_keysyms[SDLK_COMMA]= ',';
      char_keysyms[SDLK_PERIOD]= '.';
      char_keysyms[SDLK_SLASH]= '/';
      char_keysyms[SDLK_SPACE]= ' ';
      uppercase_keysyms[SDLK_a]= 'A';
      uppercase_keysyms[SDLK_b]= 'B';
      uppercase_keysyms[SDLK_c]= 'C';
      uppercase_keysyms[SDLK_d]= 'D';
      uppercase_keysyms[SDLK_e]= 'E';
      uppercase_keysyms[SDLK_f]= 'F';
      uppercase_keysyms[SDLK_g]= 'G';
      uppercase_keysyms[SDLK_h]= 'H';
      uppercase_keysyms[SDLK_i]= 'I';
      uppercase_keysyms[SDLK_j]= 'J';
      uppercase_keysyms[SDLK_k]= 'K';
      uppercase_keysyms[SDLK_l]= 'L';
      uppercase_keysyms[SDLK_m]= 'M';
      uppercase_keysyms[SDLK_n]= 'N';
      uppercase_keysyms[SDLK_o]= 'O';
      uppercase_keysyms[SDLK_p]= 'P';
      uppercase_keysyms[SDLK_q]= 'Q';
      uppercase_keysyms[SDLK_r]= 'R';
      uppercase_keysyms[SDLK_s]= 'S';
      uppercase_keysyms[SDLK_t]= 'T';
      uppercase_keysyms[SDLK_u]= 'U';
      uppercase_keysyms[SDLK_v]= 'V';
      uppercase_keysyms[SDLK_w]= 'W';
      uppercase_keysyms[SDLK_x]= 'X';
      uppercase_keysyms[SDLK_y]= 'Y';
      uppercase_keysyms[SDLK_z]= 'Z';
      uppercase_keysyms[SDLK_BACKQUOTE]= '~';
      uppercase_keysyms[SDLK_1]= '!';
      uppercase_keysyms[SDLK_2]= '@';
      uppercase_keysyms[SDLK_3]= '#';
      uppercase_keysyms[SDLK_4]= '$';
      uppercase_keysyms[SDLK_5]= '%';
      uppercase_keysyms[SDLK_6]= '^';
      uppercase_keysyms[SDLK_7]= '&';
      uppercase_keysyms[SDLK_8]= '*';
      uppercase_keysyms[SDLK_9]= '(';
      uppercase_keysyms[SDLK_0]= ')';
      uppercase_keysyms[SDLK_MINUS]= '_';
      uppercase_keysyms[SDLK_EQUALS]= '+';
      uppercase_keysyms[SDLK_LEFTBRACKET]= '{';
      uppercase_keysyms[SDLK_RIGHTBRACKET]= '}';
      uppercase_keysyms[SDLK_BACKSLASH]= '|';
      uppercase_keysyms[SDLK_SEMICOLON]= ':';
      uppercase_keysyms[SDLK_QUOTE]= '"';
      uppercase_keysyms[SDLK_COMMA]= '<';
      uppercase_keysyms[SDLK_PERIOD]= '>';
      uppercase_keysyms[SDLK_SLASH]= '?';
      uppercase_keysyms[SDLK_SPACE]= ' ';
      nfnt_font= NFNT_Load("Eyespy/Eyespy12.nfnt");
      if(nfnt_font == 0)
      {
        cerr << "Warning, could not load font:  Eyespy/Eyespy12.nfnt" << endl
          << "Command lines will not be echoed, though they will still be accepted and interpreted faithfully." << endl;
      }
      else
      {
        SDL_Color font_color= {0xff, 0xff, 0xff, 0};
        normal_font= NFNT_ToBitfont(nfnt_font, font_color);
        if(normal_font == 0)
        {
          cerr << "Warning, could not convert nfnt_font to bitfont." << endl
            << "Command lines will not be echoed, though they will still be accepted and interpreted faithfully." << endl;
        }
        else
        {
          font_inited_successfully= true;
        }
      }
      static_vars_inited= true;
    }
    string command_line;
    bool name_over= false;
    SDL_Event event;
    while(!name_over)
    {
      while(SDL_PollEvent(&event))
      {
        switch(event.type)
        {
          case SDL_KEYDOWN:
            switch(event.key.keysym.sym)
            {
              case SDLK_RETURN:
                name_over= true;
                break;
              case SDLK_BACKSPACE:
                if(command_line.size())
                {
                  /* String doesn't have pop_back(). ;_; */
                  command_line.erase(command_line.end() - 1);
                  SDL_FillRect(screen, &line_pos, 0);
                  SDL_UpdateRect(screen, line_pos.x, line_pos.y, line_pos.w, line_pos.h);
                }
                break;
              default:
                if(SDL_GetModState() & KMOD_SHIFT)
                {
                  if(uppercase_keysyms.find(event.key.keysym.sym) != uppercase_keysyms.end())
                  {
                    command_line.push_back(uppercase_keysyms[event.key.keysym.sym]);
                  }
                }
                else
                {
                  if(char_keysyms.find(event.key.keysym.sym) != char_keysyms.end())
                  {
                    command_line.push_back(char_keysyms[event.key.keysym.sym]);
                  }
                }
                break;
            }
            break;
          default:
            break;
        }
      }
      if(font_inited_successfully)
      {
        int w;
        NFNT_RenderString(normal_font, command_line.c_str(), screen, &line_pos
          , &w, 0);
        line_pos.w= w;
        SDL_UpdateRect(screen, line_pos.x, line_pos.y, line_pos.w, line_pos.h);
      }
    }
    if(command_line.empty())
    {
      throw error("No command line given.");
    }
    SDL_FillRect(screen, &line_pos, 0);
    SDL_UpdateRect(screen, line_pos.x, line_pos.y, line_pos.w, line_pos.h);
    return command_line;
  }

string get_file_name() throw(error)
  {
    string file_name= get_raw_command_line();
    for(string::iterator c= file_name.begin(); c != file_name.end(); ++c)
    {
      if(!isalnum(*c) && (*c) != '_' && (*c) != '.')
      {
        string temp("file_name contains invalid (non-alphanumeric) character:  ");
        temp+= *c;
        throw error(temp);
      }
    }
    return file_name;
  }

string get_int() throw(error)
  {
    string integer= get_raw_command_line();
    for(string::iterator c= integer.begin(); c != integer.end(); ++c)
    {
      if(!isdigit(*c) && (*c) != ' ' && (*c) != '-')
      {
        string temp("integer contains invalid (non-digit) character:  ");
        temp+= *c;
        throw error(temp);
      }
    }
    return integer;
  }

void cook_command_line(string& src, std::list<std::string>& dest)
  { /* Taken from Accelerated C++, Chapter 5 */
    typedef string::size_type string_size;
    string_size i= 0;
    // invariant: we have processed characters [original value of i, i)
    while(i != src.size())
    {
      // ignore leading blanks
      while(i != src.size() && isspace(src[i])) ++i;
      // find end of next word
      string_size j= i;
      // invariant:  none of the characters in range [original j, current j) is a space
      while(j != src.size() && !isspace(src[j])) ++j;
      // if we found some nonwhitespace characters
      if(i != j)
      {
        //copy from src starting at i and taking j - i chars
        dest.push_back(src.substr(i, j-i));
        i= j;
      }
    }
  }

void display_error(std::string& cause)
  {
    static NFNT* nfnt_font;
    static BITFONT* normal_font;
    static SDL_Rect line_pos;
    static bool font_inited_successfully= false;
    if(!font_inited_successfully)
    {
      line_pos.x= 0;
      line_pos.y= screen->h - 16;
      line_pos.w= 0;
      line_pos.h= 16;
      nfnt_font= NFNT_Load("Eyespy/Eyespy12.nfnt");
      if(nfnt_font == NULL)
      {
        cerr << "Warning, could not load font:  Eyespy/Eyespy12.nfnt" << endl
          << "Errors will not be displayed on screen.  Instead, they'll be here." << endl;
      }
      else
      {
        SDL_Color font_color= {0xff, 0, 0, 0};
        normal_font= NFNT_ToBitfont(nfnt_font, font_color);
        if(normal_font == NULL)
        {
          cerr << "Warning, could not convert nfnt_font to bitfont." << endl
              << "Errors will not be displayed on screen.  Instead, they'll be here." << endl;
        }
        else
        {
          font_inited_successfully= true;
        }
      }
    }
    if(!font_inited_successfully)
    {
      cerr << cause << endl;
      return;
    }
    int w;
    NFNT_RenderString(normal_font, cause.c_str(), screen, &line_pos, &w, 0);
    line_pos.w= w;
    SDL_UpdateRect(screen, line_pos.x, line_pos.y, line_pos.w, line_pos.h);
    bool read_error= false;
    SDL_Event ev;
    cerr << cause << endl;
    while(!read_error)
    {
      while(SDL_PollEvent(&ev))
      {
        if(ev.type == SDL_KEYDOWN)
        {
          read_error= true;
        }
      }
    }
    SDL_FillRect(screen, &line_pos, 0);
    SDL_UpdateRect(screen, line_pos.x, line_pos.y, line_pos.w, line_pos.h);
  }
