/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <map>
#include <sstream>

#include "error.hh"
#include "str_tools.hh"

using std::map;
using std::string;
using std::istringstream;

bool match_strings(string shortest_match, string longest_match, string const& culprit)
  {
    if(culprit.size() < shortest_match.size() || culprit.size() > longest_match.size())
    {
      return false;
    }
    if(culprit == shortest_match)
    {
      return true;
    }
    for(string::size_type n= 0; n < culprit.size(); ++n)
    {
      if(culprit[n] != longest_match[n])
      {
        return false;
      }
    }
    return true;
  }

bool str_to_bool(string const& str)
  {
    if(match_strings("t", "true", str))
    return true;
    return false;
  }

int str_to_int(string const& str) throw(error)
  {
    istringstream _(str);
    int ret;
    if(_ >> ret)
    {
      return ret;
    }
    string s= "Invalid int:  ";
    throw error(s + str);
  }

Uint32 str_to_uint32(string const& str) throw(error)
  {
    istringstream _(str);
    Uint32 ret;
    if((_ >> ret))
    {
      return ret;
    }
    string s= "Invalid int:  ";
    throw error(s + str);
  }

string int_to_hex_str(Uint32 i)
  {
    char digits[16]= {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c'
                      , 'd', 'e', 'f'};
    string hex_str= "0x";
    unsigned int curr_nybble= i >> 28;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 24) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 20) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 16) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 12) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 8) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= (i >> 4) & 15;
    hex_str.push_back(digits[curr_nybble]);
    curr_nybble= i & 15;
    hex_str.push_back(digits[curr_nybble]);
    return hex_str;
  }

Uint32 hex_str_to_uint32(string const& str)
  {
    map<char, unsigned int> nums;
    nums['0']= 0; nums['1']= 1; nums['2']= 2; nums['3']= 3;
    nums['4']= 4; nums['5']= 5; nums['6']= 6; nums['7']= 7;
    nums['8']= 8; nums['9']= 9; nums['a']= 10; nums['b']= 11;
    nums['c']= 12; nums['d']= 13; nums['e']= 14; nums['f']= 15;
    int d= 0;
    Uint32 ret= 0;
    for(string::const_reverse_iterator n= str.rbegin(); n != str.rend() && d < 8; ++n, ++d)
    {
      ret|= nums[*n] << (d * 4);
    }
    return ret;
  }

void operate_on_bool(string const& op_str, bool& patient)
  {
    if(match_strings("to", "toggle", op_str))
    patient= !patient;
    else
    patient= str_to_bool(op_str);
  }

size_t normalize_index(int index, size_t max_val) throw(error)
  {
    if(index < 0)
    {
      index+= max_val;
    }
    if(static_cast<size_t>(index) >= max_val)
    {
      throw error("Invalid index");
    }
    return static_cast<size_t>(index);
  }
