/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
/* I found that I needed something to take over the scut work of visualizing a line
   template for me.  That's what this is for.  It could use some improvement, but it
     works well enough for me. */
#include <iostream>
#include <string>
#include <sstream>
#include <vector>

#include "SDL.h"

#include "draw.hh"
#include "line.hh"
#include "happy_list.hh"
#include "sdl_command_line.hh"
#include "template_io.hh"

using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::istringstream;
using std::vector;

void draw_last_pixel();
void draw_current_template(Uint32 color);
void draw_unzoomed_template(basic_line& src_template, point start);
void draw_all_templates();
void draw_view();

bool cond_less_reset_bound(const int bound_breaker, int& bound);
bool cond_greater_reset_bound(const int bound_breaker, int& bound);
void cond_redralloc_template_view();
void calculate_template_dimensions(const basic_line& src, point& dest_mins, point& dest_maxs);
void reset_template_set_dimensions();
void reset_current_template();
void add_point_to_current_template(point foo);
void remove_last_point_from_current_template();

void load_template_set();
void save_template_set();

void insert_clear_template(); /* After current. */
void duplicate_current_template(); /* current_template points to the new duplicate. */
void move_to_starting_template();
void move_to_next_template();
void move_to_previous_template();
void move_to_first_template();
void move_to_last_template();

void process_click(point pos_on_screen, Uint8 button);
void process_sdlk(SDLKey foo);

int init_everything(string* template_file_name);

happy_list<basic_line> template_set;
point template_start;
point template_dimensions;
point template_mins;
point template_maxs;

point current_template_mins;
point current_template_maxs;
point last_pixel_pos;

bool show_template_set_subview= false;
int starting_template= 0;

void draw_last_pixel()
  {
    draw_pixel(last_pixel_pos, 0xff0000, false);
    update_drawn_rects();
  }

void draw_current_template(Uint32 color)
  {
    point current_pixel= template_start;
    draw_pixel(current_pixel, 0xffffff, false);
    for(vector<point>::iterator ctp= template_set.get_current_el().pixels.begin();
        ctp != template_set.get_current_el().pixels.end(); ++ctp)
    {
      current_pixel+= (*ctp);
      draw_pixel(current_pixel, color, false);
    }
    last_pixel_pos= current_pixel;
    update_drawn_rects();
  }

void draw_unzoomed_template(basic_line& src_template, point start)
  {
    SDL_Rect current_pixel= {start.x, start.y, 1, 1};
    SDL_FillRect(screen, &current_pixel, 0xffffff);
    for(vector<point>::iterator n= src_template.pixels.begin(); n != src_template.pixels.end(); ++n)
    {
      current_pixel.x+= n->x;
      current_pixel.y+= n->y;
      SDL_FillRect(screen, &current_pixel, 0xffffff);
    }
  }

void draw_all_templates()
  { /* BORKEN! */ /*
    show_template_set_subview= false;
    if(!show_template_set_subview)
    {
      return;
    }
    if(!template_set.size())
    {
      return;
    }
    unsigned int templates_per_row= screen->w / template_dimensions.x;
    unsigned int rows_of_templates= template_set.size() / templates_per_row;
    if(templates_per_row * rows_of_templates < template_set.size())
    {
      ++rows_of_templates;
    }
    point first_starting_point;
    first_starting_point.x= template_dimensions.x / 2;
    first_starting_point.y= screen->h - (template_dimensions.y * rows_of_templates);
    point current_starting_point= first_starting_point;
    for(unsigned int y= 0; y < rows_of_templates; ++y)
    {
      if((y * templates_per_row) >= template_set.size())
      {
        break;
      }
      for(unsigned int x= 0; x < templates_per_row; ++x)
      {
        if((y * templates_per_row) + x >= template_set.size())
        {
          break;
        }
        cerr << "Drawing template " << (y * templates_per_row) + x << " at "
          << current_starting_point.x << ", " << current_starting_point.y << endl;
        draw_unzoomed_template(template_set.get_indexed_el((y * templates_per_row) + x), current_starting_point);
        current_starting_point.x+= template_dimensions.x;
      }
      current_starting_point.x= first_starting_point.x;
      current_starting_point.y+= template_dimensions.y;
    }
    SDL_UpdateRect(screen, 0, 0, 0, 0); */
  }

void draw_view()
  {
    if(show_template_set_subview)
    {
      draw_all_templates();
    }
    draw_current_template(0xff0000);
  }

bool cond_less_reset_bound(const int bound_breaker, int& bound)
  {
    if(bound_breaker < bound)
    {
      bound= bound_breaker;
      return true;
    }
    return false;
  }

bool cond_greater_reset_bound(const int bound_breaker, int& bound)
  {
    if(bound_breaker > bound)
    {
      bound= bound_breaker;
      return true;
    }
    return false;
  }

void cond_redralloc_template_view()
  { /*
    if(show_template_set_subview)
    {
      if(cond_greater_reset_bound(template_maxs.x - template_mins.x, template_dimensions.x))
      {
        template_dimensions.x*= 2;
        SDL_FillRect(screen, NULL, 0);
        draw_view();
      }
      } */
  }

void calculate_template_dimensions(const basic_line& src, point& dest_mins, point& dest_maxs)
  { /*
    point current_pos;
    dest_mins.x= 0;
    dest_mins.y= 0;
    dest_maxs.x= 0;
    dest_maxs.y= 0;
    for(vector<point>::const_iterator n= src.pixels.begin(); n != src.pixels.end(); ++n)
    {
      current_pos+= *n;
      if(!cond_less_reset_bound(current_pos.x, dest_mins.x))
        cond_greater_reset_bound(current_pos.x, dest_maxs.x);
      if(!cond_less_reset_bound(current_pos.y, dest_mins.y))
        cond_greater_reset_bound(current_pos.y, dest_maxs.y);
        } */
  }

void reset_template_set_dimensions()
  {/*
    if(show_template_set_subview)
    {
      template_mins.x= 0;
      template_mins.y= 0;
      template_maxs.x= 0;
      template_maxs.y= 0;
      point cur_mins, cur_maxs;
      for(vector<basic_line>::iterator n= template_set.begin(); n != template_set.end(); ++n)
      {
        calculate_template_dimensions(*n, cur_mins, cur_maxs);
        cond_less_reset_bound(cur_mins.x, template_mins.x);
        cond_less_reset_bound(cur_mins.y, template_mins.y);
        cond_greater_reset_bound(cur_maxs.x, template_maxs.x);
        cond_greater_reset_bound(cur_maxs.y, template_maxs.y);
      }
      template_dimensions.x= template_maxs.x - template_mins.y;
      template_dimensions.y= template_maxs.y - template_mins.y;
      } */
  }

void reset_current_template()
  {
    draw_current_template(0);
    template_set.get_current_el().pixels.clear();
    current_template_mins= template_start;
    current_template_maxs= template_start;
    last_pixel_pos= template_start;
    draw_current_template(0xff0000);
    //    draw_all_templates();
  }

void add_point_to_current_template(point foo)
  {
    template_set.get_current_el().pixels.push_back(foo);
    /*
    if(show_template_set_subview)
    {
      if(cond_less_reset_bound(last_pixel_pos.x, current_template_mins.x))
      {
        if(cond_less_reset_bound(current_template_mins.x, template_mins.x))
        {
          cond_redralloc_template_view();
        }
      }
      else
      {
        if(cond_greater_reset_bound(last_pixel_pos.x, current_template_maxs.x))
        {
          if(cond_greater_reset_bound(current_template_maxs.x, template_maxs.x))
          {
            cond_redralloc_template_view();
          }
        }
      }
      } */
  }

void remove_last_point_from_current_template()
  {
    draw_pixel(last_pixel_pos, 0, false);
    last_pixel_pos-= template_set.get_current_el().pixels.back();
    template_set.get_current_el().pixels.pop_back();
  }

void load_template_set()
  {
    string file_name;
    draw_current_template(0);
    template_set.mung(); /* Sorry, you'll have to use a text editor to concatenate.
                         Don't forget to remove the extra [end template list] tag. */
    try
    {
      file_name= get_file_name();
      read_line_templates(file_name, template_set);
    }
    catch(error e)
    {
      cerr << e.cause << endl;
      return;
    }
    reset_template_set_dimensions();
    move_to_starting_template();
  }

void save_template_set()
  {
    string file_name;
    try
    {
      file_name= get_file_name();
      save_line_templates(file_name, template_set);
    }
    catch(error e)
    {
      cerr << e.cause << endl;
      return;
    }
  }

void insert_clear_template()
  {
    draw_current_template(0);
    template_set.insert(template_set.get_current_el());
    template_set.select_next_el();
    reset_current_template();
    draw_view();
  }

void duplicate_current_template()
  {
    template_set.insert(template_set.get_current_el());
    template_set.select_next_el();
    draw_view();
  }

void move_to_starting_template()
  {
    template_set.select_indexed_el(starting_template);
  }

void move_to_next_template()
  {
    draw_current_template(0);
    template_set.select_next_el();
    draw_current_template(0xff0000);
  }

void move_to_previous_template()
  {
    draw_current_template(0);
    template_set.select_previous_el();
    draw_current_template(0xff0000);
  }

void move_to_first_template()
  {
    draw_current_template(0);
    template_set.select_first_el();
    draw_current_template(0xff0000);
  }

void move_to_last_template()
  {
    draw_current_template(0);
    template_set.select_last_el();
    draw_current_template(0xff0000);
  }

void process_click(point pos_on_screen, Uint8 button)
  {
    point draw_pos= (pos_on_screen / 8) - (get_screen_offset() / 8);
    switch(button)
    {
      case SDL_BUTTON_LEFT:
        {
          point diff_from_last= draw_pos - last_pixel_pos;
          last_pixel_pos= draw_pos;
          add_point_to_current_template(diff_from_last);
          draw_last_pixel();
//           cerr << "diff_from_last:  " << diff_from_last.x << ", " << diff_from_last.y << endl;
        }
        break;
      case SDL_BUTTON_MIDDLE:
        {
          /* ADDME:  Move the center of the template view to be the point clicked. */
        }
        break;
      case SDL_BUTTON_RIGHT:
        remove_last_point_from_current_template();
        draw_current_template(0xff0000);
        break;
      default:
        cerr << "Code a case for that button if you want to use it." << endl
          << "button:  " << button << endl;
        break;
    }
  }

void process_sdlk(SDLKey foo)
  {
    switch(foo)
    {
      case SDLK_w:
        move_to_first_template();
        break;
      case SDLK_s:
        move_to_last_template();
        break;
      case SDLK_a:
        move_to_previous_template();
        break;
      case SDLK_d:
        move_to_next_template();
        break;
      case SDLK_q:
        insert_clear_template();
        break;
      case SDLK_e:
        duplicate_current_template();
        break;
      case SDLK_z:
        load_template_set();
        break;
      case SDLK_c:
        save_template_set();
        break;
      case SDLK_x:
        show_template_set_subview= !show_template_set_subview;
        break;
      default:
        break;
    }
  }

int init_everything(string& template_file_name)
  {
    try
    {
      init_drawing();
      if(!template_file_name.empty())
      {
        cerr << "Attempting to load template file." << endl;
        read_line_templates(template_file_name, template_set);
        cerr << "Partial success." << endl;
        reset_template_set_dimensions();
        cerr << "Success." << endl;
      }
      else
      {
        template_dimensions.x= 32;
        template_dimensions.y= 32;
        basic_line tmp;
        template_set.push_back(tmp);
      }
    }
    catch(error e)
    {
      cerr << e.cause << endl;
      return -1;
    }
    last_pixel_pos= template_start + get_screen_offset();
    set_zoom(8);
    return 0;
  }

void show_usage(string prog_name)
  {
    cout << "Usage:  " << prog_name << " -i file_name -v -s N" << endl
      << "-i file_name" << endl
      << "  Load template set file_name for editing." << endl
      << "-v" << endl
      << "  Toggles display of template set subview." << endl
      << "-s N" << endl
      << "  Start editing with template number N." << endl
      << "  N may be positive or negative." << endl;
  }

int main(int argc, char** argv)
  {
    string template_file_name;
    if(argc > 1)
    {
      for(int n= 1; n < argc; ++n)
      {
        if(!strcmp("-i", argv[n]))
        {
          template_file_name= argv[++n];
        }
        else if(!strcmp("-v", argv[n]))
        {
          show_template_set_subview= false;
        }
        else if(!strcmp("-s", argv[n]))
        {
          int tmp= 0;
          istringstream fod(argv[++n]);
          if(fod >> tmp)
          {
            starting_template= tmp;
          }
          else
          {
            cerr << "Invalid starting template:  " << argv[n] << endl;
            --n;
          }
        }
        else
        {
          show_usage(argv[0]);
        }
      }
    }
    if(init_everything(template_file_name) != 0)
    {
      cerr << "Error in initialization, exiting." << endl;
      SDL_Quit();
      return -1;
    }
    move_to_starting_template();
    cerr << "Successful initialization." << endl;
    draw_view();
    bool quit= false;
    while(!quit)
    {
      SDL_Event ev;
      while(SDL_PollEvent(&ev))
      {
        switch(ev.type)
        {
          case SDL_KEYDOWN:
            switch(ev.key.keysym.sym)
            {
              case SDLK_ESCAPE:
                quit= true;
                break;
              default:
                process_sdlk(ev.key.keysym.sym);
                break;
            }
            break;
          case SDL_MOUSEBUTTONDOWN:
            process_click(point(ev.button.x, ev.button.y), ev.button.button);
            break;
          case SDL_ACTIVEEVENT:
            while(!ev.active.gain)
            {
              SDL_Delay(500);
              SDL_PollEvent(&ev);
            }
            break;
          case SDL_QUIT:
            quit= true;
            break;
          default:
            break;
        }
      }
    }
    SDL_Quit();
    return 0;
  }
