/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include "str_tools.hh"
#include "template_env.hh"
#include "template_io.hh"

using std::list;
using std::string;

// #include <iostream>
// using std::cerr;
// using std::endl;

template_env::template_env(template_env& parent)
  :reset_current_after_changing_set(parent.reset_current_after_changing_set)
   , loaded_line_template_sets(parent.loaded_line_template_sets)
  {}

void template_env::init(list<string>& template_file_names)
  {
    for(list<string>::iterator n= template_file_names.begin(); n != template_file_names.end(); ++n)
    {
      load_new_template_set(*n);
    }
    for(set_of_temp_sets_t::normal_iterator n= loaded_line_template_sets.begin();
        n != loaded_line_template_sets.end(); ++n)
    {
      n->select_first_el();
    }
    loaded_line_template_sets.get_current_el().select_first_el();
//     cerr << loaded_line_template_sets.size() << " template sets loaded." << endl;
  }

basic_line& template_env::parse_template_choice_command(list<string> command) throw(error)
  {
    if(match_strings("cc", "c_current", command.front()))
    {
      command.pop_front();
      return get_current_template();
    }
    template_set_t& set_choice= loaded_line_template_sets.get_current_el();
    if(match_strings("c", "current", command.front()))
    {
      command.pop_front();
      set_choice= loaded_line_template_sets.get_current_el();
    }
    else if(match_strings("n", "next", command.front()))
    {
      command.pop_front();
      set_choice= loaded_line_template_sets.get_next_el();
    }
    else if(match_strings("p", "previous", command.front()))
    {
      command.pop_front();
      set_choice= loaded_line_template_sets.get_previous_el();
    }
    else if(match_strings("i", "indexed", command.front()))
    {
      command.pop_front();
      int index= str_to_int(command.front());
      command.pop_front();
      set_choice= loaded_line_template_sets.get_indexed_el(index);
    }
    else if(match_strings("f", "first", command.front()))
    {
      command.pop_front();
      set_choice= loaded_line_template_sets.get_first_el();
    }
    else if(match_strings("l", "last", command.front()))
    {
      command.pop_front();
      set_choice= loaded_line_template_sets.get_last_el();
    }
    else
    {
      throw error("Invalid set choice:  " + command.front());
    }
    if(match_strings("c", "current", command.front()))
    {
      command.pop_front();
      return set_choice.get_current_el();
    }
    if(match_strings("n", "next", command.front()))
    {
      command.pop_front();
      return set_choice.get_next_el();
    }
    if(match_strings("p", "previous", command.front()))
    {
      command.pop_front();
      return set_choice.get_previous_el();
    }
    if(match_strings("i", "indexed", command.front()))
    {
      command.pop_front();
      int index= str_to_int(command.front());
      command.pop_front();
      return set_choice.get_indexed_el(index);
    }
    if(match_strings("f", "first", command.front()))
    {
      command.pop_front();
      return set_choice.get_first_el();
    }
    if(match_strings("l", "last", command.front()))
    {
      command.pop_front();
      return set_choice.get_last_el();
    }
    throw error("Invalid template choice:  " + command.front());
  }

basic_line& template_env::get_current_template()
  {
    return loaded_line_template_sets.get_current_el().get_current_el();
  }

void template_env::select_first_template()
  {
    loaded_line_template_sets.get_current_el().select_first_el();
  }

void template_env::select_last_template()
  {
    loaded_line_template_sets.get_current_el().select_last_el();
  }

void template_env::select_next_template()
  {
    loaded_line_template_sets.get_current_el().select_next_el();
  }

void template_env::select_previous_template()
  {
    loaded_line_template_sets.get_current_el().select_previous_el();
  }

void template_env::select_indexed_template(int index) throw(error)
  {
    loaded_line_template_sets.get_current_el().select_indexed_el(index);
  }

void template_env::select_first_template_set()
  {
    loaded_line_template_sets.select_first_el();
  }

void template_env::select_last_template_set()
  {
    loaded_line_template_sets.select_last_el();
  }

void template_env::select_next_template_set()
  {
    loaded_line_template_sets.select_next_el();
    if(reset_current_after_changing_set)
    {
      loaded_line_template_sets.get_current_el().select_first_el();
    }
  }

void template_env::select_previous_template_set()
  {
    loaded_line_template_sets.select_previous_el();
    if(reset_current_after_changing_set)
    {
      loaded_line_template_sets.get_current_el().select_first_el();
    }
  }

void template_env::select_indexed_template_set(int index) throw(error)
  {
    loaded_line_template_sets.select_indexed_el(index);
    if(reset_current_after_changing_set)
    {
      loaded_line_template_sets.get_current_el().select_first_el();
    }
  }

void template_env::load_new_template_set(string& file_name) throw(error)
  {
    template_set_t temp;
    /* read_line_templates is the one that might throw an exception, but since
       load_new_template_set can't do anything to clean it up, catching it here would
       be inappropriate. */
    read_line_templates(file_name, temp);
    temp.select_first_el();
    loaded_line_template_sets.push_back(temp);
  }

void template_env::drop_current_template_set()
  {
    loaded_line_template_sets.remove_current_el();
    if(loaded_line_template_sets.size() < 1)
    {
      throw error("Warning:  All template sets have been dropped.  Load a new one now or this program will crash.");
    }
  }

void template_env::mung()
  {
    loaded_line_template_sets.resize(1);
  }
