/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef TEMPLATE_ENV_HH
#define TEMPLATE_ENV_HH

#include <list>
#include <string>

#include "error.hh"
#include "happy_list.hh"
#include "line.hh"

struct template_env
{
  typedef happy_list<basic_line> template_set_t;
  typedef happy_list<template_set_t> set_of_temp_sets_t;

  template_env()
    :reset_current_after_changing_set(false)
  {}

  template_env(template_env& parent);

  void init(std::list<std::string>& template_file_names);

  basic_line& parse_template_choice_command(std::list<std::string> command) throw(error);
  /* FINISHME?  Full set of command parsing functions like line_env? */
  /* If so, only after you switch to a better command structure and work out bugs
     and inconsistencies in the current structure. */
  
  basic_line& get_current_template();
  void select_first_template();
  void select_last_template();
  void select_next_template();
  void select_previous_template();
  void select_indexed_template(int index) throw(error);
  void select_first_template_set();
  void select_last_template_set();
  void select_next_template_set();
  void select_previous_template_set();
  void select_indexed_template_set(int index) throw(error);
  void load_new_template_set(std::string& file_name) throw(error);
  void drop_current_template_set();

  void mung();
  
  bool reset_current_after_changing_set;
private:
  set_of_temp_sets_t loaded_line_template_sets;

  /* There are three different things I can do for specifying the current template:
       1.  Have a single iterator that is reset to the beginning of the current set
         whenever the current set is changed.
       2.  Have a number that acts as an index and is modded by the size of the set
         when the set changes.
       3.  Have a set of iterators, each corresponding to the current template in
         each set, with another iterator into this set specifying which one is
         currently being used.
     1 has the advantage of being the simplest to implement, but it is also the least
       flexible.
     2 is actually an example of the Wrong Thing to do, as it does not have the easy
       predictability of 1 or 3, unless the user keeps in mind its current value and
       the size of the set being changed to.  It also means that for cases with
       frequent changing between different sized sets, the smaller sets will change
       the number with distressing frequency unless they are specifically avoided.
     3 is the most flexible option, though it is the hardest to implement.  It allows
       the user to easily predict which line from the chosen set will be active after
       the change (the one last selected from that set (the first if the set has not
       been used previously)).  Also, 3 can be used to imitate the behavior of 1 if
       necessary.
       While I can't think of a situation in which it would
       be used, I still want to provide this flexibility.
       happy_list makes this easy.
  */
};

#endif
