/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

#include "line.hh"
#include "template_io.hh"

using std::string;
using std::istream;
using std::ostream;
using std::cerr;
using std::endl;
using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::ostringstream;
using std::vector;

istream& read_template(istream& in_file, basic_line& dest_template) throw(error)
  {
    if(in_file)
    {
      string s;
      dest_template.pixels.clear();
      point current_point;
      while(getline(in_file, s))
      {
        if(s == "[end template]")
        {
          break;
        }
        while(s.empty() && getline(in_file, s))
        {}
        istringstream foo(s);
        if(!(foo >> current_point.x >> current_point.y))
        {
          throw error("Invalid x or y value for point:  " + s);
        }
        dest_template.pixels.push_back(current_point);
      }
    }
    return in_file;
  }

ostream& save_template(basic_line const& src_template, ostream& out_file) throw(error)
  {
    if(out_file)
    {
      for(vector<point>::const_iterator n= src_template.pixels.begin(); n != src_template.pixels.end(); ++n)
      {
        out_file << n->x << ' ' << n->y << endl;
      }
      out_file << "[end template]" << endl;
    }
    return out_file;
  }

void read_line_templates(const string& file_name, happy_list<basic_line>& dest_templates) throw(error)
  {
    ifstream template_file(file_name.c_str());
    if(!template_file)
    {
      string fisk= "Could not open template file:  ";
      fisk+= file_name;
      throw error(fisk);
    }
//     cerr << "Reading template file:  " << file_name << endl;
    string s;
    while(getline(template_file, s))
    {
      if(s == "[start template list]")
      {
        break;
      }
    }
    basic_line current_template;
    while(getline(template_file, s))
    {
      while(s.empty() && getline(template_file, s)) /* Works by short-circuit. */
      {}
      if(s == "[start template]")
      {
        try
        {
          if(!read_template(template_file, current_template))
          {
            ostringstream err_out(s);
            err_out << "Error reading template " << dest_templates.size() << endl;
            throw error(err_out.str());
          }
          dest_templates.push_back(current_template);
        }
        catch(error e)
        {
          ostringstream err_out(s);
          err_out << "Error reading template " << dest_templates.size() << ":  " << e.cause << endl;
          throw error(err_out.str());
        }
      }
      else if(s == "[end template list]")
      {
        break;
      }
    }
//     cerr << "Success." << endl;
//     cerr << dest_templates.size() << " line templates read successfully." << endl
//       << "End of list detected, closing file." << endl;
    template_file.close();
  }

void save_line_templates(const string& file_name, const happy_list<basic_line>& src_templates) throw(error)
  {
    ofstream template_file(file_name.c_str());
    if(!template_file)
    {
      throw error("Could not open template file:  " + file_name);
    }
    template_file << "[start template list]" << endl;
    for(happy_list<basic_line>::const_normal_iterator n= src_templates.begin(); n != src_templates.end(); ++n)
    {
      template_file << "[start template]" << endl;
      save_template(*n, template_file);
    }
    template_file << "[end template list]" << endl;
    template_file.close();
  }

void convert_line_to_template(line const& src, basic_line& dest) throw(error)
  {
    if(src.pixels.empty())
    {
      throw error("Blatantly refusing to convert an empty line to a template.");
    }
    point start= src.pixels.front();
    dest.pixels.clear();
    for(vector<point>::const_iterator n= src.pixels.begin() + 1; n != src.pixels.end(); ++n)
    {
      point curr_change= (*n) - start;
      dest.pixels.push_back(curr_change);
    }
  }
